XK_BackSpace = 0xFF08
XK_Tab = 0xFF09
XK_Linefeed = 0xFF0A
XK_Clear = 0xFF0B
XK_Return = 0xFF0D
XK_Pause = 0xFF13
XK_Scroll_Lock = 0xFF14
XK_Sys_Req = 0xFF15
XK_Escape = 0xFF1B
XK_Delete = 0xFFFF
XK_Multi_key = 0xFF20
XK_SingleCandidate = 0xFF3C
XK_MultipleCandidate = 0xFF3D
XK_PreviousCandidate = 0xFF3E
XK_Kanji = 0xFF21
XK_Muhenkan = 0xFF22
XK_Henkan_Mode = 0xFF23
XK_Henkan = 0xFF23
XK_Romaji = 0xFF24
XK_Hiragana = 0xFF25
XK_Katakana = 0xFF26
XK_Hiragana_Katakana = 0xFF27
XK_Zenkaku = 0xFF28
XK_Hankaku = 0xFF29
XK_Zenkaku_Hankaku = 0xFF2A
XK_Touroku = 0xFF2B
XK_Massyo = 0xFF2C
XK_Kana_Lock = 0xFF2D
XK_Kana_Shift = 0xFF2E
XK_Eisu_Shift = 0xFF2F
XK_Eisu_toggle = 0xFF30
XK_Zen_Koho = 0xFF3D
XK_Mae_Koho = 0xFF3E
XK_Home = 0xFF50
XK_Left = 0xFF51
XK_Up = 0xFF52
XK_Right = 0xFF53
XK_Down = 0xFF54
XK_Prior = 0xFF55
XK_Page_Up = 0xFF55
XK_Next = 0xFF56
XK_Page_Down = 0xFF56
XK_End = 0xFF57
XK_Begin = 0xFF58
XK_Select = 0xFF60
XK_Print = 0xFF61
XK_Execute = 0xFF62
XK_Insert = 0xFF63
XK_Undo = 0xFF65
XK_Redo = 0xFF66
XK_Menu = 0xFF67
XK_Find = 0xFF68
XK_Cancel = 0xFF69
XK_Help = 0xFF6A
XK_Break = 0xFF6B
XK_Mode_switch = 0xFF7E
XK_script_switch = 0xFF7E
XK_Num_Lock = 0xFF7F
XK_KP_Space = 0xFF80
XK_KP_Tab = 0xFF89
XK_KP_Enter = 0xFF8D
XK_KP_F1 = 0xFF91
XK_KP_F2 = 0xFF92
XK_KP_F3 = 0xFF93
XK_KP_F4 = 0xFF94
XK_KP_Home = 0xFF95
XK_KP_Left = 0xFF96
XK_KP_Up = 0xFF97
XK_KP_Right = 0xFF98
XK_KP_Down = 0xFF99
XK_KP_Prior = 0xFF9A
XK_KP_Page_Up = 0xFF9A
XK_KP_Next = 0xFF9B
XK_KP_Page_Down = 0xFF9B
XK_KP_End = 0xFF9C
XK_KP_Begin = 0xFF9D
XK_KP_Insert = 0xFF9E
XK_KP_Delete = 0xFF9F
XK_KP_Equal = 0xFFBD
XK_KP_Multiply = 0xFFAA
XK_KP_Add = 0xFFAB
XK_KP_Separator = 0xFFAC
XK_KP_Subtract = 0xFFAD
XK_KP_Decimal = 0xFFAE
XK_KP_Divide = 0xFFAF
XK_KP_0 = 0xFFB0
XK_KP_1 = 0xFFB1
XK_KP_2 = 0xFFB2
XK_KP_3 = 0xFFB3
XK_KP_4 = 0xFFB4
XK_KP_5 = 0xFFB5
XK_KP_6 = 0xFFB6
XK_KP_7 = 0xFFB7
XK_KP_8 = 0xFFB8
XK_KP_9 = 0xFFB9
XK_F1 = 0xFFBE
XK_F2 = 0xFFBF
XK_F3 = 0xFFC0
XK_F4 = 0xFFC1
XK_F5 = 0xFFC2
XK_F6 = 0xFFC3
XK_F7 = 0xFFC4
XK_F8 = 0xFFC5
XK_F9 = 0xFFC6
XK_F10 = 0xFFC7
XK_F11 = 0xFFC8
XK_L1 = 0xFFC8
XK_F12 = 0xFFC9
XK_L2 = 0xFFC9
XK_F13 = 0xFFCA
XK_L3 = 0xFFCA
XK_F14 = 0xFFCB
XK_L4 = 0xFFCB
XK_F15 = 0xFFCC
XK_L5 = 0xFFCC
XK_F16 = 0xFFCD
XK_L6 = 0xFFCD
XK_F17 = 0xFFCE
XK_L7 = 0xFFCE
XK_F18 = 0xFFCF
XK_L8 = 0xFFCF
XK_F19 = 0xFFD0
XK_L9 = 0xFFD0
XK_F20 = 0xFFD1
XK_L10 = 0xFFD1
XK_F21 = 0xFFD2
XK_R1 = 0xFFD2
XK_F22 = 0xFFD3
XK_R2 = 0xFFD3
XK_F23 = 0xFFD4
XK_R3 = 0xFFD4
XK_F24 = 0xFFD5
XK_R4 = 0xFFD5
XK_F25 = 0xFFD6
XK_R5 = 0xFFD6
XK_F26 = 0xFFD7
XK_R6 = 0xFFD7
XK_F27 = 0xFFD8
XK_R7 = 0xFFD8
XK_F28 = 0xFFD9
XK_R8 = 0xFFD9
XK_F29 = 0xFFDA
XK_R9 = 0xFFDA
XK_F30 = 0xFFDB
XK_R10 = 0xFFDB
XK_F31 = 0xFFDC
XK_R11 = 0xFFDC
XK_F32 = 0xFFDD
XK_R12 = 0xFFDD
XK_F33 = 0xFFDE
XK_R13 = 0xFFDE
XK_F34 = 0xFFDF
XK_R14 = 0xFFDF
XK_F35 = 0xFFE0
XK_R15 = 0xFFE0
XK_Shift_L = 0xFFE1
XK_Shift_R = 0xFFE2
XK_Control_L = 0xFFE3
XK_Control_R = 0xFFE4
XK_Caps_Lock = 0xFFE5
XK_Shift_Lock = 0xFFE6
XK_Meta_L = 0xFFE7
XK_Meta_R = 0xFFE8
XK_Alt_L = 0xFFE9
XK_Alt_R = 0xFFEA
XK_Super_L = 0xFFEB
XK_Super_R = 0xFFEC
XK_Hyper_L = 0xFFED
XK_Hyper_R = 0xFFEE



# Xlib.XK -- X keysym defs
#
#    Copyright (C) 2000 Peter Liljenberg <petli@ctrl-c.liu.se>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# This module defines some functions for working with X keysyms as well
# as a modular keysym definition and loading mechanism. See the keysym
# definition modules in the Xlib/keysymdef directory.

from Xlib.X import NoSymbol

def string_to_keysym(keysym):
    '''Return the (16 bit) numeric code of keysym.

    Given the name of a keysym as a string, return its numeric code.
    Don't include the 'XK_' prefix, just use the base, i.e. 'Delete'
    instead of 'XK_Delete'.'''
    return globals().get('XK_' + keysym, NoSymbol)

def load_keysym_group(group):
    '''Load all the keysyms in group.

    Given a group name such as 'latin1' or 'katakana' load the keysyms
    defined in module 'Xlib.keysymdef.group-name' into this XK module.'''
    if '.' in group:
        raise ValueError('invalid keysym group name: %s' % group)

    G = globals() #Get a reference to XK.__dict__ a.k.a. globals

    #Import just the keysyms module.
    mod = __import__('Xlib.keysymdef.%s' % group, G, locals(), [group])

    #Extract names of just the keysyms.
    keysyms = [n for n in dir(mod) if n.startswith('XK_')]

    #Copy the named keysyms into XK.__dict__
    for keysym in keysyms:
        ## k = mod.__dict__[keysym]; assert k == int(k) #probably too much.
        G[keysym] = mod.__dict__[keysym]

    #And get rid of the keysym module.
    del mod

def _load_keysyms_into_XK(mod):
    '''keysym definition modules need no longer call Xlib.XK._load_keysyms_into_XK().
    You should remove any calls to that function from your keysym modules.'''
    pass

# Always import miscellany and latin1 keysyms
load_keysym_group('miscellany')
load_keysym_group('latin1')


def keysym_to_string(keysym):
    '''Translate a keysym (16 bit number) into a python string.

    This will pass 0 to 0xff as well as XK_BackSpace, XK_Tab, XK_Clear,
    XK_Return, XK_Pause, XK_Scroll_Lock, XK_Escape, XK_Delete. For other
    values it returns None.'''

    # ISO latin 1, LSB is the code
    if keysym & 0xff00 == 0:
        return chr(keysym & 0xff)

    if keysym in [XK_BackSpace, XK_Tab, XK_Clear, XK_Return,
                  XK_Pause, XK_Scroll_Lock, XK_Escape, XK_Delete]:
        return chr(keysym & 0xff)

    # We should be able to do these things quite automatically
    # for latin2, latin3, etc, in Python 2.0 using the Unicode,
    # but that will have to wait.

    return None



#!/usr/bin/env python3

import struct
import json
import socket
import os
import re
from enum import Enum
from Xlib import display


class MessageType(Enum):
    COMMAND = 0
    GET_WORKSPACES = 1
    SUBSCRIBE = 2
    GET_OUTPUTS = 3
    GET_TREE = 4
    GET_MARKS = 5
    GET_BAR_CONFIG = 6
    GET_VERSION = 7


class Event(object):
    WORKSPACE = (1 << 0)
    OUTPUT = (1 << 1)
    MODE = (1 << 2)
    WINDOW = (1 << 3)
    BARCONFIG_UPDATE = (1 << 4)
    BINDING = (1 << 5)


class _ReplyType(dict):

    def __getattr__(self, name):
            return self[name]

    def __setattr__(self, name, value):
        self[name] = value

    def __delattr__(self, name):
        del self[name]


class CommandReply(_ReplyType):
    pass


class VersionReply(_ReplyType):
    pass


class BarConfigReply(_ReplyType):
    pass


class OutputReply(_ReplyType):
    pass


class WorkspaceReply(_ReplyType):
    pass


class WorkspaceEvent(object):

    def __init__(self, data, conn):
        self.change = data['change']
        self.current = None
        self.old = None

        if 'current' in data and data['current']:
            self.current = Con(data['current'], None, conn)

        if 'old' in data and data['old']:
            self.old = Con(data['old'], None, conn)


class GenericEvent(object):

    def __init__(self, data):
        self.change = data['change']


class WindowEvent(object):

    def __init__(self, data, conn):
        self.change = data['change']
        self.container = Con(data['container'], None, conn)


class BarconfigUpdateEvent(object):

    def __init__(self, data):
        self.id = data['id']
        self.hidden_state = data['hidden_state']
        self.mode = data['mode']


class BindingInfo(object):

    def __init__(self, data):
        self.command = data['command']
        self.mods = data['mods']
        self.input_code = data['input_code']
        self.symbol = data['symbol']
        self.input_type = data['input_type']


class BindingEvent(object):

    def __init__(self, data):
        self.change = data['change']
        self.binding = BindingInfo(data['binding'])


class _PubSub(object):

    def __init__(self, conn):
        self.conn = conn
        self._subscriptions = []

    def subscribe(self, detailed_event, handler):
        event = detailed_event.replace('-', '_')
        detail = ''

        if detailed_event.count('::') > 0:
            [event, detail] = detailed_event.split('::')

        self._subscriptions.append({'event': event, 'detail': detail,
                                    'handler': handler})

    def emit(self, event, data):
        detail = ''

        if data and hasattr(data, 'change'):
            detail = data.change

        for s in self._subscriptions:
            if s['event'] == event:
                if not s['detail'] or s['detail'] == detail:
                    if data:
                        s['handler'](self.conn, data)
                    else:
                        s['handler'](self.conn)

# this is for compatability with i3ipc-glib


class _PropsObject(object):

    def __init__(self, obj):
        object.__setattr__(self, "_obj", obj)

    def __getattribute__(self, name):
        return getattr(object.__getattribute__(self, "_obj"), name)

    def __delattr__(self, name):
        delattr(object.__getattribute__(self, "_obj"), name)

    def __setattr__(self, name, value):
        setattr(object.__getattribute__(self, "_obj"), name, value)


class Connection(object):
    MAGIC = 'i3-ipc'  # safety string for i3-ipc
    _chunk_size = 1024  # in bytes
    _timeout = 0.5  # in seconds
    _struct_header = '<%dsII' % len(MAGIC.encode('utf-8'))
    _struct_header_size = struct.calcsize(_struct_header)

    def __init__(self, socket_path=None):
        if not socket_path:
            socket_path = os.environ.get("I3SOCK")

        if not socket_path:
            d = display.Display()
            r = d.screen().root
            data = r.get_property(d.get_atom('I3_SOCKET_PATH'),
                                  d.get_atom('UTF8_STRING'), 0, 9999)

            if not data.value:
                raise Exception('could not get i3 socket path')
            socket_path = data.value

        self._pubsub = _PubSub(self)
        self.props = _PropsObject(self)
        self.subscriptions = 0
        self.socket_path = socket_path
        self.cmd_socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.cmd_socket.connect(self.socket_path)

    def _pack(self, msg_type, payload):
        """
        Packs the given message type and payload. Turns the resulting
        message into a byte string.
        """
        pb = payload.encode()
        s = struct.pack('=II', len(pb), msg_type.value)
        return self.MAGIC.encode() + s + pb

    def _unpack(self, data):
        """
        Unpacks the given byte string and parses the result from JSON.
        Returns None on failure and saves data into "self.buffer".
        """
        msg_magic, msg_length, msg_type = self._unpack_header(data)
        msg_size = self._struct_header_size + msg_length
        # XXX: Message shouldn't be any longer than the data
        return data[self._struct_header_size:msg_size].decode('utf-8')

    def _unpack_header(self, data):
        """
        Unpacks the header of given byte string.
        """
        return struct.unpack(self._struct_header,
                             data[:self._struct_header_size])

    def _ipc_recv(self, sock):
        data = sock.recv(14)

        if len(data) == 0:
            # EOF
            return '', 0

        msg_magic, msg_length, msg_type = self._unpack_header(data)
        msg_size = self._struct_header_size + msg_length
        while len(data) < msg_size:
            data += sock.recv(msg_length)
        return self._unpack(data), msg_type

    def _ipc_send(self, sock, message_type, payload):
        sock.sendall(self._pack(message_type, payload))
        data, msg_type = self._ipc_recv(sock)
        return data

    def message(self, message_type, payload):
        return self._ipc_send(self.cmd_socket, message_type, payload)

    def command(self, payload):
        data = self.message(MessageType.COMMAND, payload)
        return json.loads(data, object_hook=CommandReply)

    def get_version(self):
        data = self.message(MessageType.GET_VERSION, '')
        return json.loads(data, object_hook=VersionReply)

    def get_bar_config(self, bar_id=None):
        # default to the first bar id
        if not bar_id:
            bar_config_list = self.get_bar_config_list()
            if not bar_config_list:
                return None
            bar_id = bar_config_list[0]

        data = self.message(MessageType.GET_BAR_CONFIG, bar_id)
        return json.loads(data, object_hook=BarConfigReply)

    def get_bar_config_list(self):
        data = self.message(MessageType.GET_BAR_CONFIG, '')
        return json.loads(data)

    def get_outputs(self):
        data = self.message(MessageType.GET_OUTPUTS, '')
        return json.loads(data, object_hook=OutputReply)

    def get_workspaces(self):
        data = self.message(MessageType.GET_WORKSPACES, '')
        return json.loads(data, object_hook=WorkspaceReply)

    def get_tree(self):
        data = self.message(MessageType.GET_TREE, '')
        return Con(json.loads(data), None, self)

    def subscribe(self, events):
        events_obj = []
        if events & Event.WORKSPACE:
            events_obj.append("workspace")
        if events & Event.OUTPUT:
            events_obj.append("output")
        if events & Event.MODE:
            events_obj.append("mode")
        if events & Event.WINDOW:
            events_obj.append("window")
        if events & Event.BARCONFIG_UPDATE:
            events_obj.append("barconfig_update")
        if events & Event.BINDING:
            events_obj.append("binding")

        data = self._ipc_send(
            self.sub_socket, MessageType.SUBSCRIBE, json.dumps(events_obj))
        result = json.loads(data, object_hook=CommandReply)
        self.subscriptions |= events
        return result

    def on(self, detailed_event, handler):
        event = detailed_event.replace('-', '_')
        detail = ''

        if detailed_event.count('::') > 0:
            [event, detail] = detailed_event.split('::')

        # special case: ipc-shutdown is not in the protocol
        if event == 'ipc-shutdown':
            self._pubsub.subscribe(event, handler)
            return

        event_type = 0
        if event == "workspace":
            event_type = Event.WORKSPACE
        elif event == "output":
            event_type = Event.OUTPUT
        elif event == "mode":
            event_type = Event.MODE
        elif event == "window":
            event_type = Event.WINDOW
        elif event == "barconfig_update":
            event_type = Event.BARCONFIG_UPDATE
        elif event == "binding":
            event_type = Event.BINDING

        if not event_type:
            raise Exception('event not implemented')

        self.subscriptions |= event_type

        self._pubsub.subscribe(detailed_event, handler)

    def main(self):
        self.sub_socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.sub_socket.connect(self.socket_path)

        self.subscribe(self.subscriptions)

        while True:
            if self.sub_socket is None:
                break

            data, msg_type = self._ipc_recv(self.sub_socket)

            if len(data) == 0:
                # EOF
                self._pubsub.emit('ipc-shutdown', None)
                break

            data = json.loads(data)
            msg_type = 1 << (msg_type & 0x7f)
            event_name = ''
            event = None

            if msg_type == Event.WORKSPACE:
                event_name = 'workspace'
                event = WorkspaceEvent(data, self)
            elif msg_type == Event.OUTPUT:
                event_name = 'output'
                event = GenericEvent(data)
            elif msg_type == Event.MODE:
                event_name = 'mode'
                event = GenericEvent(data)
            elif msg_type == Event.WINDOW:
                event_name = 'window'
                event = WindowEvent(data, self)
            elif msg_type == Event.BARCONFIG_UPDATE:
                event_name = 'barconfig_update'
                event = BarconfigUpdateEvent(data)
            elif msg_type == Event.BINDING:
                event_name = 'binding'
                event = BindingEvent(data)
            else:
                # we have not implemented this event
                continue

            self._pubsub.emit(event_name, event)

    def main_quit(self):
        self.sub_socket.close()
        self.sub_socket = None


class Rect(object):

    def __init__(self, data):
        self.x = data['x']
        self.y = data['y']
        self.height = data['height']
        self.width = data['width']


class Con(object):

    def __init__(self, data, parent, conn):
        self.props = _PropsObject(self)
        self._conn = conn
        self.parent = parent

        # set simple properties
        ipc_properties = ['border', 'current_border_width', 'focused',
                          'fullscreen_mode', 'id', 'layout', 'mark', 'name',
                          'orientation', 'percent', 'type', 'urgent', 'window',
                          'num', 'scratchpad_state']
        for attr in ipc_properties:
            if attr in data:
                setattr(self, attr, data[attr])
            else:
                setattr(self, attr, None)

        # XXX this is for compatability with 4.8
        if isinstance(self.type, int):
            if self.type == 0:
                self.type = "root"
            elif self.type == 1:
                self.type = "output"
            elif self.type == 2 or self.type == 3:
                self.type = "con"
            elif self.type == 4:
                self.type = "workspace"
            elif self.type == 5:
                self.type = "dockarea"

        # set complex properties
        self.nodes = []
        for n in data['nodes']:
            self.nodes.append(Con(n, self, conn))

        self.floating_nodes = []
        for n in data['floating_nodes']:
            self.floating_nodes.append(Con(n, self, conn))

        self.window_class = None
        self.window_instance = None
        self.window_role = None
        if 'window_properties' in data:
            if 'class' in data['window_properties']:
                self.window_class = data['window_properties']['class']
            if 'instance' in data['window_properties']:
                self.window_instance = data['window_properties']['instance']
            if 'window_role' in data['window_properties']:
                self.window_role = data['window_properties']['window_role']

        self.rect = Rect(data['rect'])

    def root(self):
        if not self.parent:
            return self

        con = self.parent

        while con.parent:
            con = con.parent

        return con

    def descendents(self):
        descendents = []

        def collect_descendents(con):
            for c in con.nodes:
                descendents.append(c)
                collect_descendents(c)
            for c in con.floating_nodes:
                descendents.append(c)
                collect_descendents(c)

        collect_descendents(self)
        return descendents

    def leaves(self):
        leaves = []

        for c in self.descendents():
            if not c.nodes and c.type == "con" and c.parent.type != "dockarea":
                leaves.append(c)

        return leaves

    def command(self, command):
        self._conn.command('[con_id="{}"] {}'.format(self.id, command))

    def command_children(self, command):
        if not len(self.nodes):
            return

        commands = []
        for c in self.nodes:
            commands.append('[con_id="{}" {};'.format(self.id, command))

        self._conn.command(' '.join(commands))

    def workspaces(self):
        workspaces = []

        def collect_workspaces(con):
            if con.type == "workspace" and not con.name.startswith('__'):
                workspaces.append(con)
                return

            for c in con.nodes:
                collect_workspaces(c)

        collect_workspaces(self.root())
        return workspaces

    def find_focused(self):
        try:
            return next(c for c in self.descendents() if c.focused)
        except StopIteration:
            return None

    def find_by_id(self, id):
        try:
            return next(c for c in self.descendents() if c.id == id)
        except StopIteration:
            return None

    def find_by_window(self, window):
        try:
            return next(c for c in self.descendents() if c.window == window)
        except StopIteration:
            return None

    def find_by_role(self, pattern):
        return [c for c in self.descendents()
                if c.window_role and re.search(pattern, c.window_role)]

    def find_named(self, pattern):
        return [c for c in self.descendents()
                if c.name and re.search(pattern, c.name)]

    def find_classed(self, pattern):
        return [c for c in self.descendents()
                if c.window_class and re.search(pattern, c.window_class)]

    def find_marked(self, pattern=".*"):
        return [c for c in self.descendents()
                if c.mark and re.search(pattern, c.mark)]

    def find_fullscreen(self):
        return [c for c in self.descendents()
                if c.type == 'con' and c.fullscreen_mode]

    def workspace(self):
        ret = self.parent

        while ret:
            if ret.type == 'workspace':
                break
            ret = ret.parent

        return ret

    def scratchpad(self):
        root = self.root()

        i3con = None
        for c in root.nodes:
            if c.name == "__i3":
                i3con = c
                break

        if not i3con:
            return None

        i3con_content = None
        for c in i3con.nodes:
            if c.name == "content":
                i3con_content = c
                break

        if not i3con_content:
            return None

        scratch = None
        for c in i3con_content.nodes:
            if c.name == "__i3_scratch":
                scratch = c
                break

        return scratch



# Passphrase Dictionary
#
# Interfaces to the passphrase dictionary (the dictionary acts as the alphabet
# used to create pass phrases).
#
# Copyright (C) 2013-14 Kenneth S. Kundert and Kale Kundert

# License (fold)
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see http://www.gnu.org/licenses/.

# Imports (folds)
from .config import get_setting
from shlib import to_path
from inform import codicil, error, warn, os_error
from textwrap import dedent, wrap
import hashlib


class Dictionary:
    """Read Dictionary"""
    def __init__(self, path):
        # find the dictionary, initially look in the settings directory
        if not path.exists():
            # if not there look in install directory
            from pkg_resources import resource_filename
            path = to_path(resource_filename(__name__, 'words'))

        # open the dictionary
        try:
            contents= path.read_text()
        except OSError as err:
            error(os_error(err))
            contents = ''

        self.hash = hashlib.md5(contents.encode('utf-8')).hexdigest()
        self.words = contents.split()

DICTIONARY = Dictionary(get_setting('dictionary_file'))

# vim: set sw=4 sts=4 et:



# Xlib.support.__init__ -- support code package
#
#    Copyright (C) 2000 Peter Liljenberg <petli@ctrl-c.liu.se>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

__all__ = [
    'lock',
    'connect'
    # The platform specific modules should not be listed here
    ]



#!/usr/bin/python3

# Abraxas Password Generator
# Copyright (C) 2013-16 Kenneth S. Kundert and Kale Kundert
#
# Generates passwords and pass phrases based on stored account information.

# Imports (fold)
from abraxas import (
    PasswordGenerator, TTY_Writer, ClipboardWriter, AutotypeWriter,
    StdoutWriter, Logging)
from abraxas.prefs import (
    SEARCH_FIELDS, DEFAULT_SETTINGS_DIR, DEFAULT_ARCHIVE_FILENAME,
    BROWSERS, DEFAULT_BROWSER)
from fileutils import (
    getTail as get_tail,
    makePath as make_path,
    ShellExecute as Execute, ExecuteError)
import argparse
import sys


class CommandLine:
    def __init__(self, argv):
        """Read the Command Line"""
        self.prog_name = get_tail(argv[0])
        parser = argparse.ArgumentParser(
            add_help=False, description="Generate strong and unique password.")
        arguments = parser.add_argument_group('arguments')
        arguments.add_argument(
            'account', nargs='?', default='',
            help="Generate password specific to this account.")
        parser.add_argument(
            '-P', '--password', action='store_true',
            help="Output the password (default if nothing else is requested).")
        parser.add_argument(
            '-N', '--username', action='store_true',
            help="Output the username.")
        parser.add_argument(
            '-Q', '--question', type=int, metavar='<N>',
            default=None, help="Output security question N.")
        parser.add_argument(
            '-A', '--account-number', action='store_true',
            help="Output the account number.")
        parser.add_argument(
            '-E', '--email', action='store_true', help="Output the email.")
        parser.add_argument(
            '-U', '--url', action='store_true', help="Output the URL.")
        parser.add_argument(
            '-R', '--remarks', action='store_true',
            help="Output remarks.")
        parser.add_argument(
            '-i', '--info', action='store_true',
            help="Output everything, except the password.")
        parser.add_argument(
            '-a', '--all', action='store_true',
            help="Output everything, including the password.")
        group = parser.add_mutually_exclusive_group()
        group.add_argument(
            '-q', '--quiet', action='store_true',
            help="Disable all non-essential output.")
        group.add_argument(
            '-c', '--clipboard', action='store_true',
            help="Write output to clipboard rather than stdout.")
        group.add_argument(
            '-t', '--autotype', action='store_true',
            help=(' '.join([
                "Mimic keyboard to send output to the active window rather",
                "than stdout. In this case any command line arguments that",
                "specify what to output are ignored and the autotype entry",
                "scripts the output."])))
        parser.add_argument(
            '-f', '--find', type=str, metavar='<str>',
            help=(' '.join([
                "List any account that contains the given string",
                "in its ID or aliases."])))
        parser.add_argument(
            '-s', '--search', type=str, metavar='<str>',
            help=(' '.join([
                "List any account that contains the given string in",
                "%s, or its ID." % ', '.join(SEARCH_FIELDS)])))
        parser.add_argument(
            '-S', '--stateless', action='store_true',
            help="Do not use master password or accounts file.")
        parser.add_argument(
            '-T', '--template',
            type=str, metavar='<template>', default=None,
            help="Template to use if account is not found.")
        parser.add_argument(
            '-b', '--default-browser', action='store_true',
            help="Open account in the default browser (%s)." % DEFAULT_BROWSER)
        browsers = [
            '%s (%s)' % (k, BROWSERS[k].split()[0])
            for k in sorted(BROWSERS)
        ]
        parser.add_argument(
            '-B', '--browser', type=str, metavar='<browser>',
            help="Open account in the specified browser (choose from %s)." % (
                ', '.join(browsers)))
        parser.add_argument(
            '-n', '--notify', action='store_true',
            help="Output messages to notifier.")
        parser.add_argument(
            '-l', '--list', action='store_true',
            help=(' '.join([
                "List available master passwords and templates (only pure",
                "templates are listed, not accounts, even though accounts",
                "can be used as templates)."])))
        parser.add_argument(
            '-w', '--wait', type=float, default=60, metavar='<secs>',
            help=(' '.join([
                "Wait this long before clearing the secret",
                "(use 0 to disable)."])))
        parser.add_argument(
            '--archive', action='store_true',
            help=("Archive all the secrets to %s." % make_path(
                DEFAULT_SETTINGS_DIR, DEFAULT_ARCHIVE_FILENAME)))
        parser.add_argument(
            '--changed', action='store_true',
            help=(
                "Identify all secrets that have changed since last archived."))
        parser.add_argument(
            '-I', '--init', type=str, metavar='<GPG ID>',
            help=(' '.join([
                "Initialize the master password and account files in",
                DEFAULT_SETTINGS_DIR,
                "(but only if they do not already exist)."])))
        parser.add_argument(
            '-h', '--help',  action='store_true',
            help="Show this help message and exit.")

        args = parser.parse_args()

        # If requested, print help message and exit
        if args.help:
            parser.print_help()
            sys.exit()

        # Save all the command line arguments as attributes of self
        self.__dict__.update(args.__dict__)

    def name_as_invoked(self):
        return self.prog_name


# Main (fold)
cmd_line = CommandLine(sys.argv)
try:
    with Logging(
            argv=sys.argv, prog_name=cmd_line.name_as_invoked(),
            use_notifier=cmd_line.notify
    ) as logger:
        generator = PasswordGenerator(
            logger=logger,
            init=cmd_line.init,
            stateless=cmd_line.stateless)
        if cmd_line.init:
            logger.terminate()

        # Open the accounts file
        generator.read_accounts(cmd_line.template)

        # If requested, list the available templates and then exit
        if cmd_line.list:
            logger.display(
                "MASTER PASSWORDS:\n   " + '\n   '.join(
                    sorted(generator.master_password.password_names())))
            logger.display(
                "\nTEMPLATES:\n   " + '\n   '.join(
                    sorted(generator.all_templates())))
            logger.terminate()

        # If requested, search the account database, print results, and exit
        def print_search_results(search_term, search_func):
            to_print = []
            for acct, aliases in search_func(search_term):
                aliases = ' (%s)' % (', '.join(aliases)) if aliases else ''
                to_print += [acct + aliases]
            logger.display(search_term + ':')
            logger.display('    ' + ('\n    '.join(sorted(to_print))))

        if cmd_line.find:
            print_search_results(cmd_line.find, generator.find_accounts)
            logger.terminate()

        if cmd_line.search:
            print_search_results(
                cmd_line.search, generator.search_accounts)
            logger.terminate()

        # If requested, update or compare against archive
        if cmd_line.changed:
            generator.print_changed_secrets()
            logger.terminate()
        if cmd_line.archive:
            generator.archive_secrets()
            logger.terminate()

        # Select the requested account
        account = generator.get_account(cmd_line.account)

        # If requested, open accounts webpage in browser and then exit.
        if cmd_line.browser or cmd_line.default_browser:
            # determine which browser to use
            if cmd_line.default_browser:
                cmd = BROWSERS[DEFAULT_BROWSER]
            else:
                try:
                    cmd = BROWSERS[cmd_line.browser]
                except KeyError:
                    logger.error(
                        'Unknown browser: %s, choose from %s.' % (
                            cmd_line.browser, ', '.join(BROWSERS)
                        )
                    )

            # get the url
            urls = account.get_field('url', [])
            if type(urls) == str:
                urls = [urls]

            # run the browser
            try:
                if urls:
                    url = urls[0]  # choose first url if there is more than one
                    if '://' not in url:
                        url = 'https://' + url
                    logger.log("running '%s'" % (cmd % url))
                    Execute(cmd % url)
                    logger.terminate()
                else:
                    logger.error('url is unknown')
            except ExecuteError as err:
                logger.error(str(err))

        # Create the secrets writer
        if cmd_line.clipboard:
            writer = ClipboardWriter(generator, cmd_line.wait, logger)
        elif cmd_line.autotype:
            writer = AutotypeWriter(generator, cmd_line.wait, logger)
        elif cmd_line.quiet:
            writer = StdoutWriter(generator, cmd_line.wait, logger)
        else:
            writer = TTY_Writer(generator, cmd_line.wait, logger)

        # Process the users output requests
        if cmd_line.autotype:
            writer.write_autotype()
        else:
            if cmd_line.username or cmd_line.info or cmd_line.all:
                writer.write_account_entry('username')
            if cmd_line.account_number or cmd_line.info or cmd_line.all:
                writer.write_account_entry('account')
            if cmd_line.email or cmd_line.info or cmd_line.all:
                writer.write_account_entry('email')
            if cmd_line.url or cmd_line.info or cmd_line.all:
                writer.write_account_entry('url')
            if cmd_line.remarks or cmd_line.info or cmd_line.all:
                writer.write_account_entry('remarks')
            if cmd_line.info or cmd_line.all:
                writer.write_question()
                writer.write_unknown_entries()
            if cmd_line.question is not None:
                writer.write_answer(cmd_line.question)
            if cmd_line.password or cmd_line.all or writer.is_empty():
                writer.write_password()

        # Output everything that the user requested.
        writer.process_output()
        logger.terminate()
except KeyboardInterrupt:
    sys.exit('Killed by user')

# vim: set filetype=python sw=4 sts=4 et ai:



import pytest
from inform import os_error
from textwrap import dedent
import subprocess
import os

# set various environment variables so avendesora uses local gpg key and config
# directory rather than the users.

os.environ['HOME'] = 'home'

# test_add() {{{1
def test_add():
    try:
        result = subprocess.check_output('avendesora help add'.split())
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Add a new account.

        Usage:
            avendesora [options] add [<template>]

        Options:
            -f <file>, --file <file>
                                    Add account to specified file.

        The default template is bank. The available templates are:
        bank, command and website
    """).strip()
    assert result == bytes(expected, encoding='ascii')

# test_browse() {{{1
def test_browse():
    try:
        result = subprocess.check_output('avendesora help browse'.split())
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Open account URL in web browser.

        Usage:
            avendesora [options] browse <account> [<key>]

        Options:
            -b <browser>, --browser <browser>
                                    Open account in specified browser.

        The default browser is x. The available browsers are:
            f  firefox
            g  google-chrome
            t  torbrowser
            x  xdg-open
    """).strip()
    assert result == bytes(expected, encoding='ascii')

# test_conceal() {{{1
def test_conceal():
    try:
        result = subprocess.check_output('avendesora help conceal'.split())
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Conceal text by encoding it.

        Usage:
            avendesora [options] hide <text>
            avendesora [options] conceal <text>

        Options:
            -e <encoding>, --encoding <encoding>
                                    Encoding used when concealing information.

        Possible encodings include:
            base64

        base64 (default):
            This encoding obscures but does not encrypt the text. It can
            protect text from observers that get a quick glance of the
            encoded text, but if they are able to capture it they can easily
            decode it.
    """).strip()
    assert result == bytes(expected, encoding='ascii')

# test_edit() {{{1
def test_edit():
    try:
        result = subprocess.check_output('avendesora help edit'.split())
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Edit an account.

        Usage:
            avendesora edit <account>
    """).strip()
    assert result == bytes(expected, encoding='ascii')

# test_find() {{{1
def test_find():
    try:
        result = subprocess.check_output('avendesora help find'.split())
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Find an account.

        Find accounts whose name contains the search text.

        Usage:
            avendesora find <text>
    """).strip()
    assert result == bytes(expected, encoding='ascii')

# test_help() {{{1
def test_help():
    try:
        result = subprocess.check_output('avendesora help help'.split())
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Give information about commands or other topics.

        Usage:
            avendesora help [<topic>]
    """).strip()
    assert result == bytes(expected, encoding='ascii')

# test_init() {{{1
def test_init():
    try:
        result = subprocess.check_output('avendesora help init'.split())
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Create initial set of Avendesora files.

        Usage:
            avendesora init [--gpg-id <id>]... [options]
            avendesora initialize [--gpg-id <id>]... [options]

        Options:
            -g <id>, --gpg-id <id>  Use this ID when creating any missing encrypted files.
            -h <path>, --gpg-home <path>
                                    GPG home directory (default is ~/.gnupg).

        Initial configuration and accounts files are created only if they
        do not already exist.  Existing files are not modified.
    """).strip()
    assert result == bytes(expected, encoding='ascii')

# test_new() {{{1
def test_new():
    try:
        result = subprocess.check_output('avendesora help new'.split())
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Create new accounts file.

        Usage:
            avendesora new [--gpg-id <id>]... <name>

        Options:
            -g <id>, --gpg-id <id>  Use this ID when creating any missing encrypted files.

        Creates a new accounts file. Accounts that share the same file share
        the same master password by default and, if the file is encrypted,
        can be decrypted by the same recipients.

        Generally you would create a new accounts file for each person or
        group with which you wish to share accounts. You would use separate
        files for passwords with different security domains. For example, a
        high-value passwords might be placed in an encrypted file that would
        only be placed highly on protected computers. Conversely, low-value
        passwords might be contained in perhaps an unencrypted file that is
        found on many computers.

        Add a '.gpg' extension to <name> to encrypt the file.
    """).strip()
    assert result == bytes(expected, encoding='ascii')

# test_reveal() {{{1
def test_reveal():
    try:
        result = subprocess.check_output('avendesora help reveal'.split())
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Reveal concealed text.

        Transform concealed text to reveal its original form.

        Usage:
            avendesora [options] unhide <text>
            avendesora [options] reveal <text>

        Options:
            -e <encoding>, --encoding <encoding>
                                    Encoding used when revealing information.

        Possible encodings include:
            base64

        base64 (default):
            This encoding obscures but does not encrypt the text. It can
            protect text from observers that get a quick glance of the
            encoded text, but if they are able to capture it they can easily
            decode it.
    """).strip()
    assert result == bytes(expected, encoding='ascii')

# test_search() {{{1
def test_search():
    try:
        result = subprocess.check_output('avendesora help search'.split())
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Search accounts.

        Search for accounts whose values contain the search text.

        Usage:
            avendesora search <text>
    """).strip()
    assert result == bytes(expected, encoding='ascii')

# test_show() {{{1
def test_show():
    try:
        result = subprocess.check_output('avendesora help show'.split())
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Show an account value.

        Produce an account value. If the value is secret, it is produced only
        temporarily unless --stdout is specified.

        Usage:
            avendesora show [--stdout | --clipboard] [<account> [<field>]]
            avendesora s [--stdout | --clipboard] [<account> [<field>]]

        Options:
            -c, --clipboard         Write output to clipboard rather than stdout.
            -s, --stdout            Write output to the standard output without
                                    any annotation or protections.
    """).strip()
    assert result == bytes(expected, encoding='ascii')

# test_summary() {{{1
def test_summary():
    try:
        result = subprocess.check_output('avendesora help summary'.split())
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Display all account values.

        Show all account values.

        Usage:
            avendesora sum <account>
            avendesora summary <account>
    """).strip()
    assert result == bytes(expected, encoding='ascii')

# test_version() {{{1
def test_version():
    try:
        result = subprocess.check_output('avendesora help version'.split())
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Display Avendesora version.

        Usage:
            avendesora version
    """).strip()
    assert result == bytes(expected, encoding='ascii')

# test_overview() {{{1
def test_overview():
    try:
        result = subprocess.check_output('avendesora help overview'.split())
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Avendesora is password utility that can store your account
        information and generate account passwords and produce them from the
        command line.  It can also be configured to autotype your username
        and password into the current window so that you can log in with a
        simple keystroke.

        Avendesora is capable of generating passwords (character-based
        passcodes) or pass phrases (word-based passcodes).  Pass phrases are
        generally preferred if you have a choice, but many websites will not
        take them.  The benefit of pass phrases is that they are relatively
        easy to remember and type, and they are very secure.  The pass
        phrases generated by Avendesora generally consist of four words,
        each word is drawn from a dictionary of 10,000 words.  Thus, even if
        a bad guy knew that four lower case words were being used for your
        pass phrase, there are still 10,000,000,000,000,000 possible
        combinations for him to try (this represents a mini‐ mum entropy of
        53 bits).  Using six words results in 80 bits of entropy, which
        meets the threshold recommended by NIST for the most secure pass
        phrases. For more on this, see 'How Much Entropy is Enough' below.

        For another perspective on the attractiveness of pass phrases, see
        http://xkcd.com/936/.

        Unlike password vaults, Avendesora produces a highly unpredictable
        password from a master password and the name of the account for
        which the password is to be used. The process is completely
        repeatable. If you give the same master password and account name,
        you will get the same password. As such, the passwords do not have
        to be saved; instead they are regenerated on the fly.

        As a password generator, Avendesora provides three important
        advantages over conventional password vaults.  First, it allows
        groups of people to share access to accounts without having to
        securely share each password.  Instead, one member of the group
        creates a master password that is securely shared with the group
        once.  From then on any member of the group can create a new
        account, share the name of the account, and all members will know
        the password needed to access the account. The second advantage is
        that it opens up the possibility of using high-quality passwords for
        stealth accounts, which are accounts where you remember the name of
        the account but do not store any information about even the
        existence of the account on your computer.  With Avendesora, you
        only need to remember the name of the account and it will regenerate
        the password for you. This is perfect for your TrueCrypt hidden
        volume password.  Finally, by securely storing a small amount of
        infor‐ mation, perhaps on a piece of paper in your safe-deposit box,
        you can often recover most if not all of your passwords even if you
        somehow lose your accounts file. You can even recover passwords that
        were created after you created your backup. This is because
        Avendesora combines the master password with some easily recon‐
        structed information, such as the account name, to create the
        password. If you save the master password, the rest should be
        recoverable.

        To use it, one creates a file that contains information about each
        of his or her non-stealth accounts.  Among that information would be
        information that controls how the passwords are generated. This file
        is generally not encrypted, though you can encrypt it if you like).
        Another file is created that contains one or more master passwords.
        This file is always GPG encrypted.

        The intent is for these files to not include the passwords for your
        accounts.  Rather, the passwords are regenerated when needed from
        the account information and from the master password. This makes it
        easy to share passwords with others without having to pass the
        passwords back and forth.  It is only necessary to create a shared
        master password in advance. Then new passwords can be created on the
        fly by either party.
    """).strip()
    assert result == bytes(expected, encoding='utf8')



# Abraxas Master Password
#
# Responsible for reading and managing the data from the master password file.
#
# Copyright (C) 2013-14 Kenneth S. Kundert and Kale Kundert

# License (fold)
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see http://www.gnu.org/licenses/.

# Imports (fold)
import abraxas.secrets as secrets
import hashlib
from fileutils import (
    makePath as make_path,
    getHead as get_head,
    getExt as get_extension,
)
from abraxas.prefs import (
    DEFAULT_SETTINGS_DIR, MASTER_PASSWORD_FILENAME,
    DICTIONARY_SHA1, SECRETS_SHA1, CHARSETS_SHA1
)
from textwrap import wrap
import sys
import traceback


class _MasterPassword:
    """
    Master password class

    Responsible for reading and managing the data from the master password 
    file.
    """

    def __init__(self, path, dictionary, gpg, logger, stateless):
        self.path = path
        self.dictionary = dictionary
        self.gpg = gpg
        self.logger = logger
        self.stateless = stateless
        self.data = self._read_master_password_file()
        self.passphrase = secrets.Passphrase(
            lambda text: logger.display(text))
        self.password = secrets.Password(
            lambda text: logger.display(text))
        self._validate_assumptions()

    def _read_master_password_file(self):
        data = {
            'accounts': None,
            'passwords': {},
            'default_password': None,
            'password_overrides': {},
            'additional_master_password_files': [],
        }
        if not self.stateless:
            try:
                with open(self.path, 'rb') as f:
                    decrypted = self.gpg.decrypt_file(f)
                    if not decrypted.ok:
                        self.logger.error("%s" %
                            "%s: unable to decrypt." % (self.path),
                        )
                    code = compile(decrypted.data, self.path, 'exec')
                    exec(code, data)
            except IOError as err:
                self.logger.display(
                    'Warning: could not read master password file %s: %s.' % (
                        err.filename, err.strerror))
            except SyntaxError as err:
                traceback.print_exc(0)
                sys.exit()

        # assure that the keys on the master passwords are strings
        for ID in data.get('passwords', {}):
            if type(ID) != str:
                self.logger.error(
                    '%s: master password ID must be a string.' % ID)

        # Open additional master password files
        additional_password_files = data.get(
            'additional_master_password_files', [])
        if type(additional_password_files) == str:
            additional_password_files = [additional_password_files]
        for each in additional_password_files:
            more_data = {}
            path = make_path(get_head(self.path), each)
            if get_extension(path) in ['gpg', 'asc']:
                # File is GPG encrypted, decrypt it
                try:
                    with open(path, 'rb') as f:
                        decrypted = self.gpg.decrypt_file(f)
                        if not decrypted.ok:
                            self.logger.error("%s" %
                                "%s: unable to decrypt." % (path),
                            )
                            continue
                        code = compile(decrypted.data, path, 'exec')
                        exec(code, more_data)
                except IOError as err:
                    self.logger.display('%s: %s.  Ignored.' % (
                        err.filename, err.strerror
                    ))
                    continue
            else:
                self.logger.error(
                    "%s: must have .gpg or .asc extension" % (path))

            # Check for duplicate master passwords
            existing_names = set(data.get('passwords', {}).keys())
            new_passwords = more_data.get('passwords', {})
            new_names = set(new_passwords.keys())
            names_in_common = sorted(
                existing_names.intersection(new_names))
            if names_in_common:
                self.logger.display(
                    "%s: overrides existing password:\n    %s" % (
                        path, ',\n    '.join(sorted(names_in_common))))
            data['passwords'].update(new_passwords)

            # Check for duplicate passwords overrides
            existing_names = set(data['password_overrides'].keys())
            new_overrides = more_data.get('password_overrides', {})
            new_names = set(new_overrides.keys())
            names_in_common = sorted(
                existing_names.intersection(new_names))
            if names_in_common:
                self.logger.display(
                    "%s: overrides existing password overrides:\n    %s" % (
                        path, ',\n    '.join(sorted(names_in_common))))
            data['password_overrides'].update(new_overrides)

        return data

    def _validate_assumptions(self):
        # Check that dictionary has not changed.
        # If the master password file exists, then self.data['dict_hash'] will 
        # exist, and we will compare the current hash for the dictionary 
        # against that stored in the master password file, otherwise we will 
        # compare against the one present when the program was configured.
        self.dictionary.validate(self.data.get('dict_hash', DICTIONARY_SHA1))

        # Check that secrets.py and charset.py have not changed
        for each, sha1 in [
            ('secrets', SECRETS_SHA1),
            ('charsets', CHARSETS_SHA1)
        ]:
            path = make_path(get_head(__file__), each + '.py')
            try:
                with open(path) as f:
                    contents = f.read()
            except IOError as err:
                path = make_path(get_head(__file__), '..', each + '.py')
                try:
                    with open(path) as f:
                        contents = f.read()
                except IOError as err:
                    self.logger.error('%s: %s.' % (err.filename, err.strerror))
            hash = hashlib.sha1(contents.encode('utf-8')).hexdigest()
            # Check that file has not changed.
            # If the master password file exists, then self.data['%s_hash'] 
            # will exist, and we will compare the current hash for the file 
            # against that stored in the master password file, otherwise we 
            # will compare against the one present when the program was 
            # configured.
            if hash != self.data.get('%s_hash' % each, sha1):
                self.logger.display("Warning: '%s' has changed." % path)
                self.logger.display("    " + "\n    ".join(wrap(' '.join([
                    "This could result in passwords that are inconsistent",
                    "with those created in the past.",
                    'Update the corresponding hash in %s/%s to "%s".' % (
                        DEFAULT_SETTINGS_DIR, MASTER_PASSWORD_FILENAME, hash),
                    "Then use 'abraxas --changed' to assure that nothing has",
                    "changed."
                ]))))

    def _get_field(self, key):
        try:
            return self.data[key]
        except KeyError:
            self.logger.error("%s: cannot find '%s'" % (self.path, key))

    def get_master_password(self, account):
        """Get the master password associated with this account.

        If there is none, use the default.
        If there is no default, ask the user for a password.
        """
        passwords = self._get_field('passwords')
        default_password = self._get_field('default_password')

        # Get the master password for this account.
        if account:
            password_id = account.get_master(default_password)
        else:
            password_id = default_password
        if password_id:
            try:
                return passwords[password_id]
            except KeyError:
                self.logger.error(
                    '%s: master password not found.' % password_id)
        else:
            import getpass
            try:
                self.logger.display(
                    "Provide master password for account '%s'." % account.ID)
                master_password = getpass.getpass()
                if not master_password:
                    self.logger.display("Warning: Master password is empty.")
                return master_password
            except (EOFError, KeyboardInterrupt):
                sys.exit()

    def password_names(self):
        """Return a list that contains the name of the master passwords."""
        return self._get_field('passwords').keys()

    def generate_password(self, account, master_password=None):
        """Generate the password for the specified account

        Generally you should not need to pass in the master_password. This is
        only done for testing the stateless password generation.
        """
        try:
            return self.data['password_overrides'][account.get_id()]
        except KeyError:
            pass

        # Otherwise generate a pass phrase or a password as directed
        if not master_password:
            master_password = self.get_master_password(account)
        password_type = account.get_password_type()
        if password_type == 'words':
            return self.passphrase.generate(
                master_password, account, self.dictionary)
        elif password_type == 'chars':
            return self.password.generate(master_password, account)
        else:
            self.logger.error(
                "%s: unknown password type (expected 'words' or 'chars').")

    def generate_answer(self, account, question):
        """Generate an answer to a security question

        Question may either be the question text (a string) or it may be an
        index into the list of questions in the account (an integer).
        """
        # Configured to use only pass phrases as answers to security questions.
        # This is because people doing phone support will often simply ignore
        # those security questions that seem like gibberish to them.
        if type(question) == int:
            # question given as an index, convert it to the question text
            security_questions = account.get_security_questions()
            try:
                question = security_questions[question]
            except IndexError:
                self.logger.error(
                    'There is no security question #%s.' % question)
        master_password = self.get_master_password(account)
        answer = self.passphrase.generate(
            master_password, account, self.dictionary, question)
        return (question, answer)

# vim: set sw=4 sts=4 et:



# Abraxas Password Generator Preferences
#
# Copyright (C) 2013-14 Kenneth S. Kundert and Kale Kundert

# License (fold)
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see http://www.gnu.org/licenses/.


# Imports (fold)
from textwrap import dedent
import re


# Filenames (folds)
DEFAULT_SETTINGS_DIR = '~/.config/abraxas'
MASTER_PASSWORD_FILENAME = 'master.gpg'
DEFAULT_ACCOUNTS_FILENAME = 'accounts'
    # accounts file will be encrypted if you add .gpg or .asc extension
DICTIONARY_FILENAME = 'words'
DEFAULT_LOG_FILENAME = 'log'
    # log file will be encrypted if you add .gpg or .asc extension
DEFAULT_ARCHIVE_FILENAME = 'archive.gpg'


# Defaults (folds)
DEFAULT_TEMPLATE = "=words"
DEFAULT_AUTOTYPE = "{username}{tab}{password}{return}"
# Use absolute paths for xdotool and xsel
# Makes it harder for someone to replace them so as to expose the secrets.


# Settings (folds)
LABEL_COLOR = 'yellow'
    # choose from normal, black, red, green, yellow, blue, magenta, cyan, white
LABEL_STYLE = 'normal'
    # choose from normal, bright, reverse, dim, underline, blink, reverse,
    # invisible (these need to be implemented by underlying terminal, and some
    # are not (such a blink and dim)
INITIAL_AUTOTYPE_DELAY = 0.0
DEBUG = False
    # Turns on the logging of extra information, but may expose sensitive
    # account information in the log file.
PREFER_HTTPS = True
    # When PREFER_HTTPS is true, abraxas requires the https protocol unless
    # http is explicitly specified in the url.
    # When PREFER_HTTPS is false, abraxas allows the http protocol unless
    # https is explicitly specified in the url.


# Utility programs (folds)
XDOTOOL = '/usr/bin/xdotool'
XSEL = '/usr/bin/xsel'
GPG_BINARY = 'gpg2'
NOTIFIER_NORMAL = ['notify-send', '--urgency=low']
NOTIFIER_ERROR = ['notify-send', '--urgency=normal']

# Signatures (folds)
# These signatures must be the sha1 signatures for the corresponding files
# Regenerate them with 'sha1sum <filename>'
# These are used in creating the initial master password file.
SECRETS_SHA1 = "5d1c97a0fb699241fca5d50a7ad0508047990510"
CHARSETS_SHA1 = "dab48b2103ebde97f78cfebd15cc1e66d6af6ed0"
DICTIONARY_SHA1 = "d9aa1c08e08d6cacdf82819eeb5832429eadb95a"


# Browsers (folds)
# Associate a command with a browser key.
# The command must contain a single %s, which is replaced with URL.
BROWSERS = {
    'x': 'xdg-open %s > /dev/null', # system default browser
    'f': 'firefox -new-tab %s > /dev/null',
    'd': 'dwb %s',
    'j': 'jumanji %s',
    'c': 'google-chrome %s',
    't': 'torbrowser %s > /dev/null',
}
DEFAULT_BROWSER = 'x'


# Account Recognition (folds)
# Title Recognition
# Build up the regular expression used to recognize the various component of 
# the window title.
def labelRegex(label, regex):
    return "(?P<%s>%s)" % (label, regex)
HOST_REGEX = r'(?:[a-zA-Z0-9\-]+\.)+[a-zA-Z0-9]+'
EMAIL_REGEX = r'[a-zA-Z0-9\-]+@' + HOST_REGEX
REGEX_COMPONENTS = {
    # If you add new components, you must also add code that handles the
    # component in accounts.py.
    'title': labelRegex('title', r'.*'),
    'host': labelRegex('host', HOST_REGEX),
    'protocol': labelRegex('protocol', r'\w+'),
    'browser': labelRegex('browser', r'\w+'),
    'username': labelRegex('username', r'\w+'),
    'email': labelRegex('email', EMAIL_REGEX)}
# Hostname in Titlebar browser title regex
# Hostname in Titlebar is now named Keepass Helper
# https://addons.mozilla.org/en-US/firefox/addon/keepass-helper/?src=search
HNITB_BROWSER_TITLE_PATTERN = re.compile(
    r'(?:{title} - )?{host} \({protocol}\)(?: - {browser})?'.format(
        **REGEX_COMPONENTS
    )
)
# This is for version 3 and beyond; requires that preferences in HNINTB be set 
# to 'show the short URL' with a separator of '-'.
HNITBv3_BROWSER_TITLE_PATTERN = re.compile(
    r'(?:{title} - ){protocol}?://{host}(?: - {browser})?'.format(
        **REGEX_COMPONENTS
    )
)
# Simple browser title regex
SIMPLE_BROWSER_TITLE_PATTERN = re.compile(
    r'{title}(?: - {browser})?'.format(**REGEX_COMPONENTS))
# Recognize components of the url
URL_PATTERN = re.compile(
    r'(?:{protocol}://)?{host}(?:/.*)?'.format(**REGEX_COMPONENTS))
TITLE_PATTERNS = [
    ('hostname-in-titlebar-browser-v3', HNITBv3_BROWSER_TITLE_PATTERN),
    #('hostname-in-titlebar-browser', HNITB_BROWSER_TITLE_PATTERN),
    # You can comment out the entry above if you are not using 'Hostname in
    # Titlebar' extension to Firefox and Thunderbird
    ('simple browser title', SIMPLE_BROWSER_TITLE_PATTERN)]

# Initial master password file (folds)
MASTER_PASSWORD_FILE_INITIAL_CONTENTS = dedent('''\
    dict_hash = "%s"      # DO NOT CHANGE THIS LINE
    secrets_hash = "%s"   # DO NOT CHANGE THIS LINE
    charsets_hash = "%s"  # DO NOT CHANGE THIS LINE

    accounts_file = "%s"
    passwords = {
        'default': "%s",  # DO NOT CHANGE THIS LINE
    }
    default_password = 'default'
    password_overrides = {
        '<account>': """<place password here>""",
    }

    # vim: filetype=python sw=4 sts=4 et ai ff=unix:
''')


# Initial accounts file (folds)
ACCOUNTS_FILE_INITIAL_CONTENTS = dedent('''\
    # Account information
    #
    # Add information about each of your accounts to the accounts dictionary.
    #
    # You can use the dedent function to strip leading whitespace from
    # multi-line remarks.  You can use the character sets and exclude function
    # to create alphabets for you character-base passwords.
    #
    # Example:
    # To create an alphabet with all characters except tabs use either:
    #     'alphabet': exclude(PRINTABLE, '\\t')
    # or:
    #     'alphabet': ALPHANUMERIC + PUNCTUATION + ' '

    from textwrap import dedent
    from abraxas.charsets import (
        exclude, LOWERCASE, UPPERCASE, LETTERS, DIGITS, ALPHANUMERIC,
        HEXDIGITS, PUNCTUATION, WHITESPACE, PRINTABLE, DISTINGUISHABLE
    )

    # The desired location of the log file (use an absolute path).
    # Adding a suffix of .gpg or .asc causes the file to be encrypted 
    # (otherwise it can leak account names).
    log_file = '%s'

    # The desired location of the archive file
    # (use an absolute path, end file in .gpg)
    archive_file = '%s'

    # The GPG ID of the user (used to encrypt archive.gpg file)
    gpg_id = '%s'

    # Account Information
    # Add your account information here ...
    accounts = {
        # Templates
        # The first few are intended to be templates.  Any account can be used 
        # as a template for another account.  Those that are designated as 
        # templates (ID starts with =) cannot be used as an actual account and 
        # will not be listed in find and search results. Feel free to modify, 
        # delete, or add your own templates.  You might want to choose short 
        # names with no spaces or glob characters for those templates you plan 
        # to use from the command line. 
        "=words": {    # typically used for linux pass phrases
            'password-type': 'words',
            'num-words': 4,
            'autotype': "{password}{return}",
        },
        "=chars": {    # typically used for web passwords
            'password-type': 'chars',
            'num-chars': 12,
            'alphabet': ALPHANUMERIC + PUNCTUATION,
            'autotype': "{username}{tab}{password}{return}",
        },
        "=pin": {      # typically used for PINs
            'password-type': 'chars',
            'num-chars': 4,
            'alphabet': DIGITS,
            'autotype': "{password}{return}",
        },
        "=num": {      # typically used for PINs
            'password-type': 'chars',
            'num-chars': 8,
            'alphabet': DIGITS,
            'autotype': "{password}{return}",
        },
        "=word": {     # typically used as an alternative to a PIN
            'password-type': 'words',
            'num-words': 1,
            'autotype': "{password}{return}",
        },
        "=anum": {     # typically used for web passwords,  contains only 
                       # easily distinguished alphanumeric characters.
            'password-type': 'chars',
            'num-chars': 12,
            'alphabet': DISTINGUISHABLE,
            'autotype': "{username}{tab}{password}{return}",
        },
        "=master": {   # used to generate master passwords for abraxas
            'password-type': 'words',
            'num-words': 8,
        },
        "=extreme": {  # used in situations where there are no limits
            'password-type': 'chars',
            'num-chars': 64,
            'alphabet': exclude(PRINTABLE, '\\t'),
        },

        # Accounts
        # AAAAA
        # BBBBB
        # CCCCC
        # DDDDD
        # EEEEE
        # FFFFF
        # GGGGG
        # HHHHH
        # IIIII
        # JJJJJ
        # KKKKK
        # LLLLL
        # MMMMM
        # NNNNN
        # OOOOO
        # PPPPP
        # QQQQQ
        # RRRRR
        # SSSSS
        # TTTTT
        # UUUUU
        # VVVVV
        # WWWWW
        # XXXXX
        # YYYYY
        # ZZZZZ
        #
        # Place your accounts above. Sort them and place them below the
        # appropriate tab to make them easier to find. Here is a template that
        # shows what an account might contain. Feel free to delete this example
        # along with this comment.
        #   "<account-id>": {
        #       'username': "<username>",
        #       'account': "<account-number>",
        #       'email': "<email>",
        #       'url': "<url>",
        #       'security questions': [
        #           "<question 0>",
        #           "<question 1>",
        #           ...
        #       ],
        #       'remarks': """<remarks>""",
        #       'version': "<version>",
        #       'window': [],       # a glob string or list of glob strings 
        #                           # that are used to match window titles to 
        #                           # this account
        #       'autotype': "{username}{tab}{password}{return}",
        #       'template': "<an account id>",
        #       'master': "<a master password id>",
        #       'password-type': 'words',
        #                           # choose between "words" and "chars"
        #       'num-words': <int>, # number of words in passphrases
        #       'separator': ' ',   # separates words in passphrases
        #       'num-chars': <int>, # number of characters in passwords
        #       'alphabet': DISTINGUISHABLE
        #                           # character set used in passwords
        #                           # construct from character sets
        #       'prefix': '',       # added to the front of passwords
        #       'suffix': '',       # added to the end of passwords
        #   },
    }
    additional_accounts = []

    # vim: filetype=python sw=4 sts=4 et ai ff=unix:
''')


# Fields (folds)
# Do not change these (not user configurable)
SEARCH_FIELDS = ['username', 'account', 'email', 'url', 'remarks']
STRING_FIELDS = [
    'alphabet', 'autotype', 'email', 'master', 'prefix',
    'remarks', 'separator', 'suffix', 'template', 'type',
    'username', 'version'
]
INTEGER_FIELDS = ['num-chars', 'num-words']
LIST_FIELDS = ['security questions', 'aliases']
LIST_OR_STRING_FIELDS = ['account', 'window', 'url']
ENUM_FIELDS = {
    'password-type': ['words', 'chars']
}
ALL_FIELDS = (
    STRING_FIELDS + INTEGER_FIELDS + LIST_FIELDS + LIST_OR_STRING_FIELDS +
    [each for each in ENUM_FIELDS.keys()]
)

# vim: set sw=4 sts=4 et:



# Abraxas Password Writers
#
# Given a secret (password or passphrase) the writer is responsible for getting 
# it to the user in a reasonably secure manner.
#
# Copyright (C) 2013-14 Kenneth S. Kundert and Kale Kundert

# License (fold)
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see http://www.gnu.org/licenses/.

# Imports (fold)
from __future__ import print_function, division
import abraxas.cursor as cursor
from abraxas.prefs import (
    LABEL_COLOR, LABEL_STYLE, XDOTOOL, XSEL, ALL_FIELDS, INITIAL_AUTOTYPE_DELAY
)
from fileutils import Execute, ExecuteError
from time import sleep
import re


def indent(text, prefix='    '):
    # Indent a string
    # This should be provided by textwrap, but is not available from older 
    # versions of python.
    return '\n'.join(
        [prefix + line if line else line for line in text.split('\n')])


class Writer:
    """
    Abraxas Password Writer Base Class
    """

    # PasswordWriter is responsible for sending output to the user. It has four 
    # backends, one that writes verbosely to standard output assuming it is 
    # a TTY (TTY_Writer), one that writes quietly to standard output assuming 
    # the output is being fed another program (StdoutWriter), one that writes to 
    # the clipboard (ClipboardWriter), and one that mimics the keyboard 
    # (AutotypeWriter).  To accommodate the four backends the output is gathered 
    # up and converted into a script.  That script is interpreted by the 
    # appropriate backend to produce the output.  The script is a sequence of 
    # commands each with an argument.  Internally the script is saved as a list 
    # of tuples. The first value in the tuple is the name of the command.  
    # Several commands are supported.
    #    write_verbatim() --> ('verb', <str>)
    #        Outputs the argument verbatim.
    #    write_account_entry() --> ('interp', <label>)
    #        Interpolates information from the account file into the output.
    #        The argument is the item to be interpolated (username, url, etc.)
    #    write_unknown_entries() --> ('unknown')
    #        Interpolates any unrecognized fields in the account into the
    #        output.
    #    write_password() --> ('password')
    #        Outputs the password as a secret (the writer does its best to keep
    #        it secure).
    #    write_question() --> ('question', [<int>])
    #        Outputs security question <int>, or all the available security
    #        questions if <int> is not given. The answer is not included in the
    #        response.
    #    write_answer() --> ('answer', <int>)
    #        Outputs the answer to a security question as a secret (the output
    #        does its best to keep it secure).  The argument is the index
    #        of the question.
    #    write_sleep() --> ('sleep', <real>)
    #        Waits before continuing. The argument is the number of seconds to
    #        wait.

    def __init__(self):
        # Do not use this class directly.
        # Use one of the subclasses instead
        raise NotImplementedError

    def constructor(self, generator, wait=60, logger=None):
        """
        Arguments:
        generator (abraxas generator object)
            The password generator.
        wait (int)
            The number of seconds to wait after offering the user the password.
        logger (logger object)
            Instance of class that provides display(), log(), and error()
            methods:
                display(msg): called when a message is to be sent to the user.
                log(msg) is called when a message is only to be logged.
                error(msg): called when an error has occurred,
                            should not return.
        """
        self.wait = wait
        self.generator = generator
        self.logger = logger if logger else generator.logger
        self.script = []

    def is_empty(self):
        return not self.script

    def write_verbatim(self, text):
        """
        Specify a string to output to user when output is processed.
        """
        self.script += [('verb', text)]

    def write_account_entry(self, label):
        """
        Specify an account field to output to user when output is processed.
        """
        self.script += [('interp', label)]

    def write_unknown_entries(self):
        """
        Specify that all unrecognized account fields should be output when
        output is processed.
        """
        self.script += [('unknown',)]

    def write_password(self):
        """
        Specify that the password should be output when output is processed.
        """
        self.script += [('password',)]

    def write_question(self, num=None):
        """
        Specify that a security question should be output when output is
        processed.
        """
        self.script += [('question', num)]

    def write_answer(self, num):
        """
        Specify that the answer to a security question should be output when
        output is processed.
        """
        self.script += [('answer', num)]

    def write_sleep(self, delay):
        """
        Specify that <delay> seconds should pass before the next thing is sent
        to the output.
        """
        self.script += [('sleep', delay)]

    def write_autotype(self):
        """
        Specify that account's autotype entry should be processed and the 
        resulting output requests be placed in the writer script. Those requests 
        are honored during process_output() when the script is executed.
        """
        regex = re.compile(r'({\w+})')
        for term in regex.split(self.generator.account.get_autotype()):
            if term and term[0] == '{' and term[-1] == '}':
                cmd = term[1:-1].lower()
                if cmd in ['username', 'account', 'url', 'email', 'remarks']:
                    self.write_account_entry(cmd)
                elif cmd == 'password':
                    self.write_password()
                elif cmd == 'tab':
                    self.write_verbatim('\t')
                elif cmd == 'return':
                    self.write_verbatim('\n')
                elif cmd.startswith('sleep'):
                    cmd = cmd.split()
                    try:
                        assert cmd[0] == 'sleep'
                        assert len(cmd) == 2
                        self.write_sleep(float(cmd[1]))
                    except (AssertionError, TypeError):
                        self.logger.display("ERROR in autotype: %s" % term)
                        return
                elif cmd.startswith('question'):
                    cmd = cmd.split()
                    try:
                        assert cmd[0] == 'question'
                        assert len(cmd) == 2
                        self.write_question(int(cmd[1]))
                    except (AssertionError, TypeError, IndexError):
                        self.logger.display("ERROR in autotype: %s" % term)
                        return
                elif cmd.startswith('answer'):
                    cmd = cmd.split()
                    try:
                        assert cmd[0] == 'answer'
                        assert len(cmd) == 2
                        self.write_answer(int(cmd[1]))
                    except (AssertionError, TypeError, IndexError):
                        self.logger.display("ERROR in autotype: %s" % term)
                        return
                else:
                    self.logger.display("ERROR in autotype: %s" % term)
                    return
            else:
                if (term):
                    self.write_verbatim(term)


class TTY_Writer(Writer):
    """
    Writes output to a TTY.
    """
    def __init__(self, *args, **kwargs):
        self.constructor(*args, **kwargs)

    def process_output(self):
        """
        Process the output.

        Everything that was stashed away by the various write_ methods should
        now be sent to the user.
        """
        label_password = len(self.script) > 1

        def highlight(label, value):
            # Attach color label to a value
            return (cursor.color(
                label.upper() + ':', LABEL_COLOR, LABEL_STYLE) + ' ' + value)

        def display_secret(label, secret):
            # Send output to stdout with the labels.
            if self.wait:
                text = highlight(label, secret)
                try:
                    cursor.write(text)
                    cursor.move_back(len(text))
                    sleep(self.wait)
                    cursor.clear()
                except KeyboardInterrupt:
                    cursor.clear()
            elif label_password:
                print(highlight(label, secret))
            else:
                print(secret)

        def display_field(label, value):
            # Send field to stdout with the labels.
            if value:
                if type(value) == list:
                    print(highlight(label, '\n    '+',\n    '.join(value)))
                elif '\n' in value:
                    print(highlight(label, '\n'+indent(value.strip(), '    ')))
                else:
                    print(highlight(label, value.rstrip()))

        # Execute the script
        for action in self.script:
            if action[0] == 'interp':
                display_field(
                    action[1],
                    self.generator.account.get_field(action[1]))
            elif action[0] == 'unknown':
                fields = sorted(
                    set(self.generator.account.get_data().keys()) -
                    set(ALL_FIELDS))
                for field in fields:
                    display_field(
                        field, self.generator.account.get_field(field))
            elif action[0] == 'password':
                display_secret(
                    'PASSWORD',
                    self.generator.generate_password()
                )
            elif action[0] == 'question':
                questions = self.generator.account.get_field(
                    'security questions')
                if questions:
                    if action[1] is None:
                        for index, question in enumerate(questions):
                            print(highlight('QUESTION %d' % index, question))
                    else:
                        try:
                            print(highlight(
                                'QUESTION %d' % action[1],
                                questions[action[1]]))
                        except IndexError:
                            print(highlight(
                                'QUESTION %d' % action[1],
                                '<not available>'))
            elif action[0] == 'answer':
                question, answer = self.generator.generate_answer(action[1])
                if answer:
                    display_secret(question, answer)
            else:
                raise NotImplementedError
        self.logger.log('Writing to stdout.')


class ClipboardWriter(Writer):
    """
    Writes output to the system clipboard.
    """
    def __init__(self, *args, **kwargs):
        self.constructor(*args, **kwargs)

    def process_output(self):
        """
        Process the output.

        Everything that was stashed away by the various write_ methods should
        now be sent to the user. The labels are suppressed.
        """
        lines = []

        # Execute the script
        for action in self.script:
            if action[0] == 'interp':
                value = self.generator.account.get_field(action[1])
                if type(value) == list:
                    lines += ["%s: %s" % (action[1], ', '.join(value))]
                elif value:
                    lines += ["%s: %s" % (action[1], value.rstrip())]
            elif action[0] == 'unknown':
                fields = sorted(
                    set(self.generator.account.get_data().keys()) -
                    set(ALL_FIELDS))
                for field in fields:
                    value = self.generator.account.get_field(field)
                    if type(value) == list:
                        lines += ["%s: %s" % (field, ', '.join(value))]
                    elif value:
                        lines += ["%s: %s" % (field, value.rstrip())]
            elif action[0] == 'password':
                lines += [self.generator.generate_password()]
            elif action[0] == 'question':
                questions = self.generator.account.get_field(
                    'security questions')
                if questions:
                    if action[1] is None:
                        for index, question in enumerate(questions):
                            lines += ['question %d: %s' % (index, question)]
                    else:
                        try:
                            lines += ['question %d: %s' % (
                                action[1], questions[action[1]])]
                        except IndexError:
                            lines += [
                                'question %d: <not available>' % action[1]]
            elif action[0] == 'answer':
                question, answer = self.generator.generate_answer(action[1])
                if answer:
                    lines += [answer]
            else:
                raise NotImplementedError
        text = '\n'.join(lines)
        self.logger.log('Writing to clipboard.')

        # Use 'xsel' to put the information on the clipboard.
        # This represents a vulnerability, if someone were to replace xsel they
        # could steal my passwords. This is why I use an absolute path. I tried
        # to access the clipboard directly using GTK but I cannot get the code
        # to work.
        try:
            Execute([XSEL, '-b', '-i'], stdin=text)
        except ExecuteError as err:
            self.logger.error(str(err))
        try:
            sleep(self.wait)
        except KeyboardInterrupt:
            pass
        try:
            Execute([XSEL, '-b', '-c'])
        except ExecuteError as err:
            self.logger.error(str(err))

        # Use Gobject Introspection (GTK) to put the information on the
        # clipboard (for some reason I cannot get this to work).
        #try:
        #    from gi.repository import Gtk, Gdk
        #
        #    clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
        #    clipboard.set_text(text, len(text))
        #    clipboard.store()
        #    sleep(self.wait)
        #    clipboard.clear()
        #except ImportError:
        #    error('Clipboard is not supported.')


class AutotypeWriter(Writer):
    """
    Writes output via autotype (it appears to be coming from the keyboard).
    """
    def __init__(self, *args, **kwargs):
        self.constructor(*args, **kwargs)

    def process_output(self):
        """
        Process the output.

        Everything that was stashed away by the various write_ methods should
        now be sent to the user.
        """

        def autotype(text):
            # Use 'xdotool' to mimic the keyboard.
            # A dollar sign in the argument to 'type' is treated as an 
            # environment variable, so it must also be separated out and sent 
            # as a explicit 'key' stroke.
            # type' must be the last action on a xdotool command line, so 
            # special characters (dollar sign) following text demands another 
            # invocation of xdotool.
            # It is desirable to pump the actions into standard in rather than 
            # place them on the command line so no part of the password is 
            # visible using ps, however this mode seems flaky in xdotool. So 
            # I have compromised and only send the individual 'key' strokes on 
            # the command line and send the 'type' text through stdin. Still 
            # seems flaky though, especially with Firefox.
            regex = re.compile(r'([\n$]+)')

            def add_action(action, arg):
                actions.append((action, arg))

            # split string so that special characters are isolated
            actions = []
            segments = regex.split(text)
            for segment in segments:
                for char in segment:
                    if char == '\n':
                        add_action('key', 'Return')
                    elif char == '$':
                        add_action('key', 'dollar')
                    else:
                        add_action('type', segment)
                        break

            def run_xdotool(args, text=None):
                try:
                    if args:
                        Execute(
                            [XDOTOOL, 'getactivewindow'] + args,
                        )
                    if text:
                        Execute(
                            [XDOTOOL, '-'],
                            stdin="getactivewindow type '%s'" % text
                        )
                except ExecuteError as err:
                    self.logger.error(str(err))

            # gather keys until 'type' is found, and then output gathered keys 
            # and type string all at once; this minimizes the number of times 
            # xdotool must be called.
            args = []
            for action, arg in actions:
                if action == 'type':
                    run_xdotool(args, arg)
                    args = []
                else:
                    args += [action, arg]
            run_xdotool(args)

        # Execute the script
        text = []
        scrubbed = []
        sleep(INITIAL_AUTOTYPE_DELAY)
        for action in self.script:
            if action[0] == 'verb':
                text += [action[1]]
                scrubbed += [action[1]]
            elif action[0] == 'sleep':
                autotype(''.join(text))
                text = []
                sleep(action[1])
                scrubbed += ['<sleep %s>' % action[1]]
            elif action[0] == 'interp':
                value = self.generator.account.get_field(action[1])
                if type(value) == list:
                    value = ', '.join(value)
                elif value:
                    value = value.rstrip()
                else:
                    value = '<%s unknown>' % action[1]
                text += [value]
                scrubbed += [value]
            elif action[0] == 'password':
                text += [self.generator.generate_password()]
                scrubbed += ['<<password>>']
            elif action[0] == 'question':
                questions = self.generator.account.get_field(
                    'security questions')
                if questions:
                    if action[1] is None:
                        for index, question in enumerate(questions):
                            value = 'question %d: %s' % (index, question)
                    else:
                        try:
                            value = 'question %d: %s' % (
                                action[1], questions[action[1]])
                        except IndexError:
                            value = 'question %d: <not available>' % action[1]
                    text += [value]
                scrubbed += [value]
            elif action[0] == 'answer':
                question, answer = self.generator.generate_answer(action[1])
                if answer:
                    text += [answer]
                    scrubbed += ["<<answer to '%s'>>" % question]
            else:
                raise NotImplementedError
        self.logger.log('Autotyping "%s".' % ''.join(scrubbed))
        autotype(''.join(text))


class StdoutWriter(Writer):
    """
    Writes output to the standard output. Suppresses everything except the 
    password.
    """
    def __init__(self, *args, **kwargs):
        self.constructor(*args, **kwargs)

    def process_output(self):
        """
        Process the output.

        Everything that was stashed away by the various write_ methods should
        now be sent to the user.

        Writes only essential information (the secret) to stdout.  This is meant 
        to facilitate scripting with abraxas.
        """

        for action in self.script:
            if action[0] == 'interp':
                action = self.generator.account.get_field(action[1])
                if action:
                    print(action)

            elif action[0] == 'password':
                print(self.generator.generate_password())

            elif action[0] == 'answer':
                print(self.generator.generate_answer(action[1])[1])

            else:
                pass

        self.logger.log(
            'Writing quietly to stdout.  Some output may be suppressed.')

# vim: set sw=4 sts=4 et:



# Xlib.Xutil -- ICCCM definitions and similar stuff
#
#    Copyright (C) 2000 Peter Liljenberg <petli@ctrl-c.liu.se>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


NoValue = 0x0000
XValue = 0x0001
YValue = 0x0002
WidthValue = 0x0004
HeightValue = 0x0008
AllValues = 0x000F
XNegative = 0x0010
YNegative = 0x0020
USPosition = (1 << 0)
USSize = (1 << 1)
PPosition = (1 << 2)
PSize = (1 << 3)
PMinSize = (1 << 4)
PMaxSize = (1 << 5)
PResizeInc = (1 << 6)
PAspect = (1 << 7)
PBaseSize = (1 << 8)
PWinGravity = (1 << 9)
PAllHints = (PPosition|PSize|PMinSize|PMaxSize|PResizeInc|PAspect)
InputHint = (1 << 0)
StateHint = (1 << 1)
IconPixmapHint = (1 << 2)
IconWindowHint = (1 << 3)
IconPositionHint = (1 << 4)
IconMaskHint = (1 << 5)
WindowGroupHint = (1 << 6)
MessageHint = (1 << 7)
UrgencyHint	= (1 << 8)
AllHints = (InputHint|StateHint|IconPixmapHint|IconWindowHint|
            IconPositionHint|IconMaskHint|WindowGroupHint|MessageHint|
            UrgencyHint)
WithdrawnState = 0
NormalState = 1
IconicState = 3
DontCareState = 0
ZoomState = 2
InactiveState = 4
RectangleOut = 0
RectangleIn = 1
RectanglePart = 2
VisualNoMask = 0x0
VisualIDMask = 0x1
VisualScreenMask = 0x2
VisualDepthMask = 0x4
VisualClassMask = 0x8
VisualRedMaskMask = 0x10
VisualGreenMaskMask = 0x20
VisualBlueMaskMask = 0x40
VisualColormapSizeMask = 0x80
VisualBitsPerRGBMask = 0x100
VisualAllMask = 0x1FF
ReleaseByFreeingColormap = 1
BitmapSuccess = 0
BitmapOpenFailed = 1
BitmapFileInvalid = 2
BitmapNoMemory = 3
XCSUCCESS = 0
XCNOMEM = 1
XCNOENT = 2



XK_emspace = 0xaa1
XK_enspace = 0xaa2
XK_em3space = 0xaa3
XK_em4space = 0xaa4
XK_digitspace = 0xaa5
XK_punctspace = 0xaa6
XK_thinspace = 0xaa7
XK_hairspace = 0xaa8
XK_emdash = 0xaa9
XK_endash = 0xaaa
XK_signifblank = 0xaac
XK_ellipsis = 0xaae
XK_doubbaselinedot = 0xaaf
XK_onethird = 0xab0
XK_twothirds = 0xab1
XK_onefifth = 0xab2
XK_twofifths = 0xab3
XK_threefifths = 0xab4
XK_fourfifths = 0xab5
XK_onesixth = 0xab6
XK_fivesixths = 0xab7
XK_careof = 0xab8
XK_figdash = 0xabb
XK_leftanglebracket = 0xabc
XK_decimalpoint = 0xabd
XK_rightanglebracket = 0xabe
XK_marker = 0xabf
XK_oneeighth = 0xac3
XK_threeeighths = 0xac4
XK_fiveeighths = 0xac5
XK_seveneighths = 0xac6
XK_trademark = 0xac9
XK_signaturemark = 0xaca
XK_trademarkincircle = 0xacb
XK_leftopentriangle = 0xacc
XK_rightopentriangle = 0xacd
XK_emopencircle = 0xace
XK_emopenrectangle = 0xacf
XK_leftsinglequotemark = 0xad0
XK_rightsinglequotemark = 0xad1
XK_leftdoublequotemark = 0xad2
XK_rightdoublequotemark = 0xad3
XK_prescription = 0xad4
XK_minutes = 0xad6
XK_seconds = 0xad7
XK_latincross = 0xad9
XK_hexagram = 0xada
XK_filledrectbullet = 0xadb
XK_filledlefttribullet = 0xadc
XK_filledrighttribullet = 0xadd
XK_emfilledcircle = 0xade
XK_emfilledrect = 0xadf
XK_enopencircbullet = 0xae0
XK_enopensquarebullet = 0xae1
XK_openrectbullet = 0xae2
XK_opentribulletup = 0xae3
XK_opentribulletdown = 0xae4
XK_openstar = 0xae5
XK_enfilledcircbullet = 0xae6
XK_enfilledsqbullet = 0xae7
XK_filledtribulletup = 0xae8
XK_filledtribulletdown = 0xae9
XK_leftpointer = 0xaea
XK_rightpointer = 0xaeb
XK_club = 0xaec
XK_diamond = 0xaed
XK_heart = 0xaee
XK_maltesecross = 0xaf0
XK_dagger = 0xaf1
XK_doubledagger = 0xaf2
XK_checkmark = 0xaf3
XK_ballotcross = 0xaf4
XK_musicalsharp = 0xaf5
XK_musicalflat = 0xaf6
XK_malesymbol = 0xaf7
XK_femalesymbol = 0xaf8
XK_telephone = 0xaf9
XK_telephonerecorder = 0xafa
XK_phonographcopyright = 0xafb
XK_caret = 0xafc
XK_singlelowquotemark = 0xafd
XK_doublelowquotemark = 0xafe
XK_cursor = 0xaff



import pytest
import os

# set various environment variables so avendesora uses local gpg key and config
# directory rather than the users.

os.environ['HOME'] = 'home'

from avendesora import PasswordGenerator, Error

def test_login():
    try:
        pw = PasswordGenerator()
        account = pw.get_account('login')
        passphrase = account.get_value()
    except Error as err:
        passphrase = str(err)
    assert passphrase == 'archangel fiesta ripen tracksuit autobahn integer'

def test_mybank_accounts_checking():
    try:
        pw = PasswordGenerator()
        account = pw.get_account('mybank')
        checking = account.get_value('accounts.checking')
    except Error as err:
        checking = str(err)
    assert checking == '12345678'

def test_mybank_accounts_checking2():
    try:
        pw = PasswordGenerator()
        account = pw.get_account('mybank')
        checking = account.get_field('accounts', 'checking')
    except Error as err:
        checking = str(err)
    assert str(checking) == '12345678'

def test_alertscc():
    try:
        pw = PasswordGenerator()
        account = pw.get_account('alertscc')
        password = account.get_value()
    except Error as err:
        password = str(err)
    assert password == 'vGnKdofWRFLT'

def test_alertscc_question1():
    try:
        pw = PasswordGenerator()
        account = pw.get_account('alertscc')
        answer = account.get_value(1)
    except Error as err:
        answer = str(err)
    assert answer == 'clipping knight guest'

def test_alertscc_question2():
    try:
        pw = PasswordGenerator()
        account = pw.get_account('alertscc')
        answer = account.get_field('questions', 2)
    except Error as err:
        answer = str(err)
    assert str(answer) == 'mistrust tumor bonfire'



# Help
# Output a help topic.
#
# To Do:  Output the obscured text within a class instantiation. So rather than
# just outputting obscured text, output 'Hidden('obscured-text'). Also, accept
# obscured text within its class and try to decode it using the class name, so
# if GPG("ciphertext") then it will use GPG to decode it even if user does not
# provide the encoding.
#
# Also implement script class that uses a personal password. Call this Encrypt
# or PersonalEncrypt.

# License {{{1
# Copyright (C) 2016 Kenneth S. Kundert
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see http://www.gnu.org/licenses/.


# Imports {{{1
from .conceal import Conceal
from .command import Command
from .config import get_setting
from .utilities import pager, two_columns
from inform import error, output
from textwrap import dedent

# HelpMessage base class {{{1
class HelpMessage:
    # get_name() {{{2
    @classmethod
    def get_name(cls):
        try:
            return cls.name.lower()
        except AttributeError:
            # consider converting lower to upper case transitions in __name__ to
            # dashes.
            return cls.__name__.lower()

    # topics {{{2
    @classmethod
    def topics(cls):
        for sub in cls.__subclasses__():
            yield sub

    # show {{{2
    @classmethod
    def show(cls, name=None):
        if name:
            name = name.lower()
            command = Command.find(name)
            if command:
                return pager(command.help())
            for topic in cls.topics():
               if name == topic.get_name():
                   return pager(topic.help())
            error('topic not found.', culprit=name)
        else:
            cls.help()

    # summarize {{{2
    @classmethod
    def summarize(cls, width=16):
        summaries = []
        for topic in sorted(cls.topics(), key=lambda topic: topic.get_name()):
            summaries.append(two_columns(topic.get_name(), topic.DESCRIPTION))
        return '\n'.join(summaries)

    # help {{{2
    @classmethod
    def help(cls):
        output('Available commands:')
        output(Command.summarize())

        output('\nAvailable topics:')
        output(cls.summarize())


# Overview class {{{1
class Overview(HelpMessage):
    DESCRIPTION = "overview of Avendesora"

    @staticmethod
    def help():
        text = dedent("""
            Avendesora is password utility that can store your account
            information and generate account passwords and produce them from the
            command line.  It can also be configured to autotype your username
            and password into the current window so that you can log in with a
            simple keystroke.

            Avendesora is capable of generating passwords (character-based
            passcodes) or pass phrases (word-based passcodes).  Pass phrases are
            generally preferred if you have a choice, but many websites will not
            take them.  The benefit of pass phrases is that they are relatively
            easy to remember and type, and they are very secure.  The pass
            phrases generated by Avendesora generally consist of four words,
            each word is drawn from a dictionary of 10,000 words.  Thus, even if
            a bad guy knew that four lower case words were being used for your
            pass phrase, there are still 10,000,000,000,000,000 possible
            combinations for him to try (this represents a mini‐ mum entropy of
            53 bits).  Using six words results in 80 bits of entropy, which
            meets the threshold recommended by NIST for the most secure pass
            phrases. For more on this, see 'How Much Entropy is Enough' below.

            For another perspective on the attractiveness of pass phrases, see
            http://xkcd.com/936/.

            Unlike password vaults, Avendesora produces a highly unpredictable
            password from a master password and the name of the account for
            which the password is to be used. The process is completely
            repeatable. If you give the same master password and account name,
            you will get the same password. As such, the passwords do not have
            to be saved; instead they are regenerated on the fly.

            As a password generator, Avendesora provides three important
            advantages over conventional password vaults.  First, it allows
            groups of people to share access to accounts without having to
            securely share each password.  Instead, one member of the group
            creates a master password that is securely shared with the group
            once.  From then on any member of the group can create a new
            account, share the name of the account, and all members will know
            the password needed to access the account. The second advantage is
            that it opens up the possibility of using high-quality passwords for
            stealth accounts, which are accounts where you remember the name of
            the account but do not store any information about even the
            existence of the account on your computer.  With Avendesora, you
            only need to remember the name of the account and it will regenerate
            the password for you. This is perfect for your TrueCrypt hidden
            volume password.  Finally, by securely storing a small amount of
            infor‐ mation, perhaps on a piece of paper in your safe-deposit box,
            you can often recover most if not all of your passwords even if you
            somehow lose your accounts file. You can even recover passwords that
            were created after you created your backup. This is because
            Avendesora combines the master password with some easily recon‐
            structed information, such as the account name, to create the
            password. If you save the master password, the rest should be
            recoverable.

            To use it, one creates a file that contains information about each
            of his or her non-stealth accounts.  Among that information would be
            information that controls how the passwords are generated. This file
            is generally not encrypted, though you can encrypt it if you like).
            Another file is created that contains one or more master passwords.
            This file is always GPG encrypted.

            The intent is for these files to not include the passwords for your
            accounts.  Rather, the passwords are regenerated when needed from
            the account information and from the master password. This makes it
            easy to share passwords with others without having to pass the
            passwords back and forth.  It is only necessary to create a shared
            master password in advance. Then new passwords can be created on the
            fly by either party.
        """).strip()
        return text





# Account
# API for an account

# License {{{1
# Copyright (C) 2016 Kenneth S. Kundert
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see http://www.gnu.org/licenses/.


# Imports {{{1
from .browsers import StandardBrowser
from .conceal import Conceal
from .config import get_setting
from .preferences import TOOL_FIELDS
from .recognize import Recognizer
from .secrets import Secret
from inform import Color, Error, is_collection, log, output, warn
from textwrap import indent, dedent
from urllib.parse import urlparse
import re
import sys


# Globals {{{1
VECTOR_PATTERN = re.compile(r'\A(\w+)\[(\w+)\]\Z')
LabelColor = Color(
    color=get_setting('label_color'),
    scheme=get_setting('color_scheme'),
    enable=Color.isTTY()
)

# Utilities {{{1
# items {{{2
def items(collection):
    # iterate through a collection returning key, value pairs in sorted order
    # collection may be a dictionary, an array, or a scalar
    try:
        # assume a dictionary
        iterator = sorted(collection.keys())
    except AttributeError:
        # it is not a dictionary
        if is_collection(collection):
            iterator = range(len(collection))  # it is an array
        else:
            yield None, collection             # it is a scalar
            return

    for key in iterator:
        yield key, collection[key]

# Account Class {{{1
class Account:
    # all_accounts() {{{2
    @classmethod
    def all_accounts(cls):
        for sub in cls.__subclasses__():
            yield sub
            for each in sub.all_accounts():
                yield each

    # fields() {{{2
    @classmethod
    def fields(cls):
        for key, value in cls.__dict__.items():
            if not key.startswith('_'):
                yield key, value

    # get_name() {{{2
    @classmethod
    def get_name(cls):
        try:
            return cls.NAME
        except AttributeError:
            # consider converting lower to upper case transitions in __name__ to
            # dashes.
            return cls.__name__.lower()

    # matches_exactly() {{{2
    @classmethod
    def matches_exactly(cls, account):
        if account == cls.get_name():
            return True
        try:
            if account in cls.aliases:
                return True
        except AttributeError:
            pass
        return False

    # id_contains() {{{2
    @classmethod
    def id_contains(cls, target):
        target = target.lower()
        if target in cls.get_name().lower():
            return True
        try:
            for alias in cls.aliases:
                if target in alias.lower():
                    return True
        except AttributeError:
            pass
        return False

    # account_contains() {{{2
    @classmethod
    def account_contains(cls, target):
        if cls.id_contains(target):
            return True
        target = target.lower()
        for key, value in cls.fields():
            if key in TOOL_FIELDS:
                continue
            try:
                if is_collection(value):
                    for k, v in items(value):
                        if target in v.lower():
                            return True
                elif target in value.lower():
                    return True
            except AttributeError:
                # is not a string, and so 
                pass
        return False

    # recognize() {{{2
    @classmethod
    def recognize(cls, data):
        # try the specified recognizers
        discovery = getattr(cls, 'discovery', ())
        for recognizer in discovery:
            if isinstance(recognizer, Recognizer):
                script = recognizer.match(data, cls)
                if script:
                    return script
        if discovery:
            return

        # If no recognizers specified, just check the url
        try:
            urls = cls.get_field('url')
        except Error:
            return
        for _, url in items(urls):
            components = urlparse(url)
            protocol = components.scheme
            host = components.netloc
            if host == data.get('host'):
                if (
                    protocol != data.get('protocol') and
                    data['protocol'] in get_setting('required_protocols')
                ):
                    msg = 'url matches, but uses wrong protocol.'
                    notify(msg)
                    error(msg, culprit=account.get_name())
                else:
                    return True

    # initialize() {{{2
    @classmethod
    def initialize(cls):
        log('initializing', cls.get_name())
        try:
            if cls.master.is_secure():
                if not cls._file_info.encrypted:
                    warn(
                        'high value master password not contained in encrypted',
                        'account file.', culprit=cls.get_name()
                    )
        except AttributeError as err:
            pass

    # values() {{{2
    @classmethod
    def values(cls):
        for key in sorted(cls.__dict__):
            if not key.startswith('_'):
                yield key, cls.__dict__[key]

    # get_field() {{{2
    @classmethod
    def get_field(cls, name, key=None, default=False):
        """Get Field Value

        Return value from the account given a field name and key.
        """
        value = cls.__dict__.get(name)
        if value is None:
            if default is False:
                raise Error('not found.', culprit=cls.combine_name(name, key))
            else:
                return default

        if key is None:
            if is_collection(value):
                raise Error(
                    'composite value found, need key.',
                    culprit=name,
                    is_collection=True,
                    collection = value
                )
        else:
            try:
                value = value[key]
            except (IndexError, KeyError, TypeError):
                raise Error('not found.', culprit=cls.combine_name(name, key))

        # generate the value if needed
        try:
            value.generate(key, cls)
        except AttributeError as err:
            pass
        return value

    # is_secret() {{{2
    @classmethod
    def is_secret(cls, name, key=None):
        value = cls.__dict__.get(name)
        if key is None:
            return hasattr(value, 'generate')
        else:
            try:
                return hasattr(value[key], 'generate')
            except (IndexError, KeyError, TypeError):
                raise Error('not found.', culprit=cls.combine_name(name, key))

    # split_name() {{{2
    @classmethod
    def split_name(cls, name):
        # Account fields can either be scalars or composites (vectors or
        # dictionaries). This function takes a string (name) that the user
        # provides to specify which account value they wish and splits it into a
        # field name and a key.  If the field is a scalar, the key will be None.
        # Users request a value using one of the following forms:
        #     True: use default name
        #     field: scalar value (key=None)
        #     index: questions (field->'questions', key=index)
        #     field[index] or field/index: for vector value
        #     field[key] or field/key: for dictionary value

        if name is True or not name:
            name = cls.get_field('default', default=None)
        if not name:
            name = get_setting('default_field')

        # convert dashes to underscores
        name = str(name).replace('-', '_')

        # If name is an integer, treat it as number of security question.
        try:
            return get_setting('default_vector_field'), int(name)
        except ValueError:
            pass

        # Split name if given in the form: name/key
        try:
            name, key = name.split('.')
            try:
                return name, int(key)
            except ValueError:
                return name, key
        except ValueError:
            pass

        # Split name if given in the form: name[key]
        match = VECTOR_PATTERN.match(name)
        if match:
            # vector name using 'name[key]' syntax
            name, key = match.groups()
            try:
                return name, int(key)
            except ValueError:
                return name, key

        # Must be scalar name
        return name, None

    # combine_name() {{{2
    @classmethod
    def combine_name(cls, name, key=None):
        # Inverse of split_name().

        # convert underscores to dashes
        #name = name.replace('_', '-')

        if key is None:
            return name
        else:
            return '%s.%s' % (name, key)

    # get_value() {{{2
    @classmethod
    def get_value(cls, name=None):
        """Get Account Value

        Return value from the account given a user friendly identifier. User
        friendly identifiers include:
            name: scalar value
            name.key or name[key]:
                member of a dict or array
                dict if key is string, array if key is number
        """
        value = cls.get_field(*cls.split_name(name))
        if isinstance(value, Secret) or isinstance(value, Conceal):
            value = str(value)
        return value

    # write_summary() {{{2
    @classmethod
    def write_summary(cls):
        # present all account values that are not explicitly secret to the user

        def fmt_field(key, value='', level=0):
            if '\n' in value:
                value = indent(dedent(value), get_setting('indent')).strip('\n')
                sep = '\n'
            elif value:
                sep = ' '
            else:
                sep = ''
            key = str(key).upper().replace('_', ' ')
            leader = level*get_setting('indent')
            return indent(LabelColor(key + ':') + sep + value, leader)

        def reveal(name, key=None):
            return "<reveal with 'avendesora show %s %s'>" % (
                cls.get_name(), cls.combine_name(name, key)
            )

        def extract_collection(name, collection):
            lines = [fmt_field(key)]
            for k, v in items(collection):
                if hasattr(v, 'generate'):
                    # is a secret, get description if available
                    try:
                        v = '%s %s' % (v.get_description(), reveal(name, k))
                    except AttributeError:
                        v = reveal(name, k)
                lines.append(fmt_field(k, v, level=1))
            return lines

        # preload list with the names associated with this account
        names = [cls.get_name()]
        if hasattr(cls, 'aliases'):
            names += cls.aliases
        lines = [fmt_field('names', ', '.join(names))]

        for key, value in cls.values():
            if key in TOOL_FIELDS:
                # is an Avendesora field
                pass
            elif is_collection(value):
                lines += extract_collection(key, value)
            elif hasattr(value, 'generate'):
                lines.append(fmt_field(key, reveal(key)))
            else:
                lines.append(fmt_field(key, value))
        output(*lines, sep='\n')

    # open_browser() {{{2
    @classmethod
    def open_browser(cls, name):
        browser = cls.get_field('browser', default=None)
        if browser is None or is_str(browser):
            browser = StandardBrowser(name)
        url = cls.get_field('url', default=None)
        browser.run(url)



# Xlib.xobject.colormap -- colormap object
#
#    Copyright (C) 2000 Peter Liljenberg <petli@ctrl-c.liu.se>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import re
import string

from Xlib import error
from Xlib.protocol import request
from Xlib.xobject import resource

rgb_res = [
    re.compile(r'\Argb:([0-9a-fA-F]{1,4})/([0-9a-fA-F]{1,4})/([0-9a-fA-F]{1,4})\Z'),
    re.compile(r'\A#([0-9a-fA-F])([0-9a-fA-F])([0-9a-fA-F])\Z'),
    re.compile(r'\A#([0-9a-fA-F][0-9a-fA-F])([0-9a-fA-F][0-9a-fA-F])([0-9a-fA-F][0-9a-fA-F])\Z'),
    re.compile(r'\A#([0-9a-fA-F][0-9a-fA-F][0-9a-fA-F])([0-9a-fA-F][0-9a-fA-F][0-9a-fA-F])([0-9a-fA-F][0-9a-fA-F][0-9a-fA-F])\Z'),
    re.compile(r'\A#([0-9a-fA-F][0-9a-fA-F][0-9a-fA-F][0-9a-fA-F])([0-9a-fA-F][0-9a-fA-F][0-9a-fA-F][0-9a-fA-F])([0-9a-fA-F][0-9a-fA-F][0-9a-fA-F][0-9a-fA-F])\Z'),
    ]

class Colormap(resource.Resource):
    __colormap__ = resource.Resource.__resource__

    def free(self, onerror = None):
        request.FreeColormap(display = self.display,
                             onerror = onerror,
                             cmap = self.id)

        self.display.free_resource_id(self.id)

    def copy_colormap_and_free(self, scr_cmap):
        mid = self.display.allocate_resource_id()
        request.CopyColormapAndFree(display = self.display,
                                    mid = mid,
                                    src_cmap = src_cmap)

        cls = self.display.get_resource_class('colormap', Colormap)
        return cls(self.display, mid, owner = 1)

    def install_colormap(self, onerror = None):
        request.InstallColormap(display = self.display,
                                onerror = onerror,
                                cmap = self.id)

    def uninstall_colormap(self, onerror = None):
        request.UninstallColormap(display = self.display,
                                  onerror = onerror,
                                  cmap = self.id)

    def alloc_color(self, red, green, blue):
        return request.AllocColor(display = self.display,
                                  cmap = self.id,
                                  red = red,
                                  green = green,
                                  blue = blue)

    def alloc_named_color(self, name):
        for r in rgb_res:
            m = r.match(name)
            if m:
                rs = m.group(1)
                r = string.atoi(rs + '0' * (4 - len(rs)), 16)

                gs = m.group(2)
                g = string.atoi(gs + '0' * (4 - len(gs)), 16)

                bs = m.group(3)
                b = string.atoi(bs + '0' * (4 - len(bs)), 16)

                return self.alloc_color(r, g, b)

        try:
            return request.AllocNamedColor(display = self.display,
                                           cmap = self.id,
                                           name = name)
        except error.BadName:
            return None

    def alloc_color_cells(self, contiguous, colors, planes):
        return request.AllocColorCells(display = self.display,
                                       contiguous = contiguous,
                                       cmap = self.id,
                                       colors = colors,
                                       planes = planes)

    def alloc_color_planes(self, contiguous, colors, red, green, blue):
        return request.AllocColorPlanes(display = self.display,
                                        contiguous = contiguous,
                                        cmap = self.id,
                                        colors = colors,
                                        red = red,
                                        green = green,
                                        blue = blue)

    def free_colors(self, pixels, plane_mask, onerror = None):
        request.FreeColors(display = self.display,
                           onerror = onerror,
                           cmap = self.id,
                           plane_mask = plane_mask,
                           pixels = pixels)

    def store_colors(self, items, onerror = None):
        request.StoreColors(display = self.display,
                            onerror = onerror,
                            cmap = self.id,
                            items = items)

    def store_named_color(self, name, pixel, flags, onerror = None):
        request.StoreNamedColor(display = self.display,
                                onerror = onerror,
                                flags = flags,
                                cmap = self.id,
                                pixel = pixel,
                                name = name)

    def query_colors(self, pixels):
        r = request.QueryColors(display = self.display,
                                cmap = self.id,
                                pixels = pixels)
        return r.colors

    def lookup_color(self, name):
        return request.LookupColor(display = self.display,
                                   cmap = self.id,
                                   name = name)



import pytest
from inform import os_error
from textwrap import dedent
import subprocess
import os

# set various environment variables so avendesora uses local gpg key and config
# directory rather than the users.

os.environ['HOME'] = 'home'

# test_mybank() {{{1
def test_mybank():
    try:
        result = subprocess.check_output('avendesora show --stdout mybank'.split())
    except OSError as err:
        result = os_error(err)
    assert result == b'&5U9eZ+iT84T\n'

# test_mb() {{{1
def test_mb():
    try:
        result = subprocess.check_output('avendesora show --stdout mb'.split())
    except OSError as err:
        result = os_error(err)
    assert result == b'&5U9eZ+iT84T\n'

# test_mb_checking() {{{1
def test_mb_checking():
    try:
        result = subprocess.check_output('avendesora show --stdout mb accounts.checking'.split())
    except OSError as err:
        result = os_error(err)
    assert result == b'12345678\n'

# test_mb_savings() {{{1
def test_mb_savings():
    try:
        result = subprocess.check_output('avendesora show --stdout mb accounts[savings]'.split())
    except OSError as err:
        result = os_error(err)
    assert result == b'23456789\n'

# test_alertscc() {{{1
def test_alertscc():
    try:
        result = subprocess.check_output('avendesora show --stdout alertscc'.split())
    except OSError as err:
        result = os_error(err)
    assert result == b'vGnKdofWRFLT\n'

# test_scc() {{{1
def test_scc():
    try:
        result = subprocess.check_output('avendesora show --stdout scc'.split())
    except OSError as err:
        result = os_error(err)
    assert result == b'vGnKdofWRFLT\n'

# test_scc_account() {{{1
def test_scc_account():
    try:
        result = subprocess.check_output('avendesora show --stdout scc account'.split())
    except OSError as err:
        result = os_error(err)
    assert result == b'123456-7890\n'

# test_scc_birthdate() {{{1
def test_scc_birthdate():
    try:
        result = subprocess.check_output('avendesora show --stdout scc birthdate'.split())
    except OSError as err:
        result = os_error(err)
    assert result == b'1978-10-18\n'

# test_scc_email() {{{1
def test_scc_email():
    try:
        result = subprocess.check_output('avendesora show --stdout scc email'.split())
    except OSError as err:
        result = os_error(err)
    assert result == b'pizzaman@pizza.com\n'

# test_scc_q0() {{{1
def test_scc_q0():
    try:
        result = subprocess.check_output('avendesora show --stdout scc 0'.split())
    except OSError as err:
        result = os_error(err)
    assert result == b'sniff flashy putsch\n'

# test_scc_q1() {{{1
def test_scc_q1():
    try:
        result = subprocess.check_output('avendesora show --stdout scc questions.1'.split())
    except OSError as err:
        result = os_error(err)
    assert result == b'clipping knight guest\n'

# test_scc_q2() {{{1
def test_scc_q2():
    try:
        result = subprocess.check_output('avendesora show --stdout scc questions[2]'.split())
    except OSError as err:
        result = os_error(err)
    assert result == b'mistrust tumor bonfire\n'

# test_login() {{{1
def test_login():
    try:
        result = subprocess.check_output('avendesora show --stdout login'.split())
    except OSError as err:
        result = os_error(err)
    assert result == b'archangel fiesta ripen tracksuit autobahn integer\n'

# test_summary() {{{1
def test_summary():
    try:
        result = subprocess.check_output('avendesora summary mb'.split())
    except OSError as err:
        result = os_error(err)
    expected = dedent("""\
        NAMES: mybank, mb
        ACCOUNTS:
            CHECKING: <reveal with 'avendesora show mybank accounts.checking'>
            CREDITCARD: <reveal with 'avendesora show mybank accounts.creditcard'>
            SAVINGS: <reveal with 'avendesora show mybank accounts.savings'>
        CUSTOMER SERVICE: 1-866-229-6633
        EMAIL: pizzaman@pizza.com
        PASSCODE: <reveal with 'avendesora show mybank passcode'>
        PIN: <reveal with 'avendesora show mybank pin'>
        QUESTIONS:
            0: What city were you born in? <reveal with 'avendesora show mybank questions.0'>
            1: What street did you grow up on? <reveal with 'avendesora show mybank questions.1'>
            2: What was your childhood nickname? <reveal with 'avendesora show mybank questions.2'>
        URL: https://mb.com
        USERNAME: pizzaman
        VERBAL: <reveal with 'avendesora show mybank verbal'>
    """)
    assert result == bytes(expected, encoding='ascii')

# test_find() {{{1
def test_find():
    try:
        result = subprocess.check_output('avendesora find bank'.split())
    except OSError as err:
        result = os_error(err)
    expected = dedent("""\
        bank:
            mybank (mb)
    """)
    assert result == bytes(expected, encoding='ascii')

# test_search() {{{1
def test_search():
    try:
        result = subprocess.check_output('avendesora search pizza'.split())
    except OSError as err:
        result = os_error(err)
    expected = dedent("""\
        pizza:
            alertscc (scc)
            mybank (mb)
    """)
    assert result == bytes(expected, encoding='ascii')

# test_reveal() {{{1
def test_reveal():
    try:
        result = subprocess.check_output('avendesora reveal MTIzNDU2Nzg='.split())
    except OSError as err:
        result = os_error(err)
    assert result == b'12345678\n'

# test_conceal() {{{1
def test_conceal():
    try:
        result = subprocess.check_output('avendesora conceal 12345678'.split())
    except OSError as err:
        result = os_error(err)
    assert result == b'MTIzNDU2Nzg=\n'



import os, errno
from fnmatch import fnmatch

"""Various utilities for interacting with files and directories."""

# Globbing Utilities {{{1
def getAll(pattern='*'):
    """
    Get all items that match a particular glob pattern.
    Supports *, ? and [] globbing (including [!]), but does not support {}
    """
    import glob
    pattern = os.path.expanduser(pattern)
    pattern = os.path.expandvars(pattern)
    return glob.glob(pattern)

def getFiles(pattern='*'):
    """
    Get all files that match a particular glob pattern.
    """
    return [each for each in getAll(pattern) if os.path.isfile(each)]

def getDirs(pattern='*'):
    """
    Get all directories that match a particular glob pattern.
    """
    return [each for each in getAll(pattern) if os.path.isdir(each)]

def getFilesRecursively(path, acceptCriteria = None, rejectCriteria = None, exclude = None):
    """
    Returns a generator that iterates through all the files contained in a
    directory hierarchy.  Accept and reject criteria are glob strings, or lists
    of glob strings. For a file to be returned its name must not match any of
    the reject criteria if any are given, and it must match one of the accept
    criteria, if any are given.  If no criteria are given, all files are
    returned. Exclude is a file or directory or a list of files or directories
    to exclude. Each is specified relative from the current working directory.
    """
    if type(acceptCriteria) == str:
        acceptCriteria = [acceptCriteria]
    if type(rejectCriteria) == str:
        rejectCriteria = [rejectCriteria]
    def yieldFile(filename):
        filename = getTail(filename)
        if acceptCriteria != None:
            if rejectCriteria != None:
                for criterion in rejectCriteria:
                    if fnmatch(filename, criterion):
                        return False
            for criterion in acceptCriteria:
                if fnmatch(filename, criterion):
                    return True
            return False
        else:
            if rejectCriteria != None:
                for criterion in rejectCriteria:
                    if fnmatch(filename, criterion):
                        return False
            return True

    def prepExcludes(exclude):
        if not exclude:
            return []
        if type(exclude) == str:
            exclude = [exclude]
        excludes = []
        for each in exclude:
            excludes += [splitPath(each)]
        return excludes

    def skip(path, excludes):
        for each in excludes:
            if splitPath(path)[0:len(each)] == each:
                return True
        return False

    if isFile(path):
        if yieldFile(path):
            yield path
    else:
        excludes = prepExcludes(exclude)
        for path, subdirs, files in os.walk(path):
            for file in files:
                filename = makePath(path, file)
                if skip(filename, excludes):
                    continue
                if yieldFile(filename):
                    yield filename

# Type Utilities {{{1
def isFile(path):
    """
    Determine whether path corresponds to a file.
    """
    return os.path.isfile(path)

def isDir(path):
    """
    Determine whether path corresponds to a directory.
    """
    return os.path.isdir(path)

def isLink(path):
    """
    Determine whether path corresponds to a symbolic link.
    """
    return os.path.islink(path)

def exists(path):
    """
    Determine whether path corresponds to a file or directory.
    """
    return os.path.exists(path)

# File Type Utilities {{{2
def fileIsReadable(filepath):
    """
    Determine whether file exists and is readable by user.
    """
    return os.path.isfile(filepath) and os.access(filepath, os.R_OK)

def fileIsExecutable(filepath):
    """
    Determine whether file exists and is executable by user.
    """
    return os.path.isfile(filepath) and os.access(filepath, os.X_OK)

def fileIsWritable(filepath):
    """
    If filepath exists, determine whether it is writable by user.
    If not, determine if directory is writable by user.
    """
    if os.path.isfile(filepath) and os.access(filepath, os.W_OK):
        return True
    else:
        path = os.path.split(filepath)
        dirpath = os.path.join(*path[:-1])
        return os.path.isdir(dirpath) and os.access(filepath, os.W_OK)

# Directory Utilities {{{2
def dirIsReadable(dirpath):
    """
    Determine whether directory exists and is readable by user.
    """
    return os.path.isdir(dirpath) and os.access(dirpath, os.R_OK)

def dirIsWritable(dirpath):
    """
    Determine whether directory exists and is executable by user.
    """
    return os.path.isdir(dirpath) and os.access(dirpath, os.W_OK)


# Path Utilities {{{1
# Joins arguments into a filesystem path
def makePath(*args):
    """
    Join the arguments together into a filesystem path.
    """
    if len(args):
        return os.path.join(*args)
    else:
        return ''

# Splits path at directory boundaries into its component pieces
def splitPath(path):
    """
    Split the path at directory boundaries.
    """
    return path.split(os.path.sep)

# Return normalized path
def normPath(path):
    """
    Convert to a normalized path (remove redundant separators and up-level references).
    """
    return os.path.normpath(path)

# Return absolute path
def absPath(path):
    """
    Convert to an absolute path.
    """
    return os.path.abspath(normPath(path))

# Return relative path
def relPath(path, start = None):
    """
    Convert to an relative path.
    """
    if start:
        return os.path.relpath(normPath(path), start)
    else:
        return os.path.relpath(normPath(path))

# Perform common path expansions (user, envvars)
def expandPath(path):
    """
    Expand initial ~ and any variables in a path.
    """
    path = os.path.expandvars(path)
    path = os.path.expanduser(path)
    path = os.path.normpath(path)
    return path

# Access components of a path {{{2
# Head (dir/file.ext ==> dir)
def getHead(path):
    """
    Return head of path: dir/file.ext ==> dir
    """
    return os.path.split(path)[0]

# Tail (dir/file.ext ==> file.ext)
def getTail(path):
    """
    Return tail of path: dir/file.ext ==> file.ext
    """
    return os.path.split(path)[1]

# Root (dir/file.ext ==> dir/file)
def getRoot(path):
    """
    Return root of path: dir/file.ext ==> dir/file
    """
    return os.path.splitext(path)[0]

# Extension (dir/file.ext ==> ext)
def getExt(path):
    """
    Return root of path: dir/file.ext ==> ext
    """
    ext = os.path.splitext(path)[1]
    return ext[1:] if ext else ''

# Copy, Move, Remove Utilities {{{1
def copy(src, dest):
    """
    Copy either a file or a directory.
    """
    import shutil
    destpath = splitPath(dest)[0:-1]
    if destpath:
        destpath = makePath(*destpath)
        if not exists(destpath):
            os.makedirs(destpath)
    try:
        if isDir(src):
            shutil.copytree(src, dest, symlinks=True)
        else:
            shutil.copy2(src, dest)
    except (IOError, OSError) as err:
        exit("%s: %s." % (err.filename, err.strerror))
    except shutil.Error as err:
        exit(["%s to %s: %s." % arg for arg in err.args])

def move(src, dest):
    """
    Move either a file or a directory.
    """
    import shutil
    try:
        shutil.move(src, dest)
    except (IOError, OSError) as err:
        exit("%s: %s." % (err.filename, err.strerror))
    except shutil.Error as err:
        exit("Cannot rename %s to %s: %s." % (src, dest, ', '.join(err.args)))

def remove(path, exitUponError=True):
    """
    Remove either a file or a directory.
    """
    # if exists(path): # do not test for existence, this will causes misdirected symlinks to be ignored
    try:
        if isDir(path):
            import shutil
            shutil.rmtree(path)
        else:
            os.remove(path)
    except (IOError, OSError) as err:
        # don't complain if the file never existed
        if err.errno != errno.ENOENT:
            if exitUponError:
                exit("%s: %s." % (err.filename, err.strerror))
            else:
                raise

def makeLink(src, dest):
    """
    Create a symbolic link.
    """
    try:
        os.symlink(src, dest)
    except (IOError, OSError) as err:
        exit("%s: %s." % (dest, err.strerror))

def mkdir(path):
    """
    Create a directory if it does not exist. If it does, return without complaint.
    """
    try:
        os.makedirs(path)
    except (IOError, OSError) as err:
        if err.errno != errno.EEXIST:
            exit("%s: %s." % (err.filename, err.strerror))

# Execute Utilities {{{1
class ExecuteError(Exception):
    def __init__(self, cmd, error, filename=None, showCmd=False):
        self.cmd = cmd if type(cmd) is str else ' '.join(cmd)
        self.filename = filename
        self.error = error
        self.showCmd = showCmd

    def __str__(self):
        if self.showCmd:
            cmd = (
                self.filename
                if self.filename
                else (
                    self.cmd if self.showCmd == 'full' else self.cmd.split()[0]
                )
            )
            return "%s: %s" % (cmd, self.error)
        else:
            return "%s" % self.error

class Execute():
    def __init__(
        self, cmd, accept=(0,), stdin=None, stdout=True, stderr=True, wait=True, 
        shell=False, showCmd=False
    ):
        """
        Execute a command and capture its output.

        cmd may be a list of strings (preferred) or a string. Ex: ['rm' '-f' 
        'core'] or 'rm -f core'.

        Raise an ExecuteError if return status is not in accept unless accept
        is set to True. By default, only a status of 0 is accepted.

        If stdin is None, no connection is made to the standard input, otherwise 
        stdin is expected to be a string and that string is sent to stdin.

        If stdout / stderr is true, stdout / stderr is captured and made 
        available from self.stdout / self.stderr.

        If wait is true, the run method does not return until the process ends.  
        In this case run() returns the status. Otherwise it return None and 
        instead calling wait() waits for the process to end and returns the 
        status.  Once wait() returns, either by calling constructor with 
        wait=True, or by calling wait(), the status is also available from 
        self.status,

        The default is to not use a shell to execute a command (safer).
        """
        self.cmd = cmd
        self.accept = accept
        self.save_stdout = stdout
        self.save_stderr = stderr
        self.wait_for_termination = wait
        self.showCmd = showCmd
        self._run(stdin, shell)

    def _run(self, stdin, shell):
        import subprocess
        streams = {}
        if stdin is not None:
            streams['stdin'] = subprocess.PIPE
        if self.save_stdout:
            streams['stdout'] = subprocess.PIPE
        if self.save_stderr:
            streams['stderr'] = subprocess.PIPE
        try:
            process = subprocess.Popen(
                self.cmd, shell=shell, **streams
            )
        except (IOError, OSError) as err:
            raise ExecuteError(self.cmd, err.filename, err.strerror)
        if stdin is not None:
            process.stdin.write(stdin.encode('utf-8'))
            process.stdin.close()
        self.pid = process.pid
        self.process = process
        if self.wait_for_termination:
            return self.wait()

    def wait(self):
        if self.save_stdout:
            self.stdout = self.process.stdout.read().decode('utf-8')
        else:
             self.stderr = None
        if self.save_stderr:
            self.stderr = self.process.stderr.read().decode('utf-8')
        else:
             self.stderr = None
        self.status = self.process.wait()
        if self.process.stdout:
            self.process.stdout.close()
        if self.process.stderr:
            self.process.stderr.close()
        if self.accept is not True and self.status not in self.accept:
            if self.stderr:
                raise ExecuteError(self.cmd, self.stderr, showCmd=self.showCmd)
            else:
                raise ExecuteError(
                    self.cmd,
                    "unexpected exit status (%d)." % self.status,
                    showCmd=self.showCmd)
        return self.status


class ShellExecute(Execute):
    def __init__(
        self, cmd, accept=(0,), stdin=None, stdout=True, stderr=True, wait=True, 
        shell=True, showCmd=False
    ):
        """
        Execute a command in a shell and capture its output

        This class is the same as Execute, except that by default it runs the 
        given command in a shell, which is less safe but often more convenient.
        """
        self.cmd = cmd
        self.accept = accept
        self.save_stdout = stdout
        self.save_stderr = stderr
        self.wait_for_termination = wait
        self.showCmd = showCmd
        self._run(stdin, True)


def execute(cmd, accept=(0,), stdin=None, shell=False):
    """
    Execute a command without capturing its output

    Raise an ExecuteError if return status is not in accept unless accept
    is set to True. By default, only a status of 0 is accepted. The default is 
    to not use a shell to execute a command (safer).
    If stdin is None, no connection is made to the standard input, otherwise 
    stdin is expected to be a string.
    """
    import subprocess
    streams = {'stdin': subprocess.PIPE} if stdin is not None else {}
    try:
        process = subprocess.Popen(cmd, shell=shell, **streams)
    except (IOError, OSError) as err:
        raise ExecuteError(cmd, err.filename, err.strerror)
    if stdin is not None:
        process.stdin.write(stdin.encode('utf-8'))
        process.stdin.close()
    status = process.wait()
    if accept is not True and status not in accept:
        raise ExecuteError(
            cmd,
            "unexpected exit status (%d)." % status,
            showCmd='brief'
        )
    return status

def shellExecute(cmd, accept=(0,), stdin=None, shell=True):
    """
    Execute a command without capturing its output

    Raise an ExecuteError if return status is not in accept unless accept
    is set to True. By default, only a status of 0 is accepted. The default is 
    to use a shell to execute a command (more convenient).
    """
    return execute(cmd, accept, stdin, shell=True)


def background(cmd, stdin=None, shell=False):
    """
    Execute a command in the background without capturing its output.

    If stdin is None, no connection is made to the standard input, otherwise 
    stdin is expected to be a string.
    """
    import subprocess
    streams = {'stdin': subprocess.PIPE} if stdin is not None else {}
    try:
        process = subprocess.Popen(cmd, shell=shell, **streams)
    except (IOError, OSError) as err:
        raise ExecuteError(cmd, err.filename, err.strerror)
    if stdin is not None:
        process.stdin.write(stdin.encode('utf-8'))
        process.stdin.close()
    return process.pid

def shellBackground(cmd, stdin=None, shell=True):
    """
    Execute a command in the background without capturing its output.

    If stdin is None, no connection is made to the standard input, otherwise 
    stdin is expected to be a string.

    The default is to use a shell to execute a command (more convenient).
    """
    return execute(cmd, stdin, shell=True)


def which(name, flags=os.X_OK):
    """Search PATH for executable files with the given name.

    Arguments:
    name (str): The name for which to search.
    flags (int): Arguments to os.access.

    A list of the full paths to files found, in the order in which they were
    found.
    """
    result = []
    path = os.environ.get('PATH', '')
    for p in path.split(os.pathsep):
        p = os.path.join(p, name)
        if os.access(p, flags):
            result.append(p)
    return result

# vim: set sw=4 sts=4 et:



# Avendesora Password Generator Preferences
#
# Copyright (C) 2016 Kenneth S. Kundert

# License {{{1
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see http://www.gnu.org/licenses/.

# Imports {{{1
from textwrap import dedent
import re
from appdirs import user_config_dir

# Non-Config Settings {{{1
# These value can be accessed with get_settings,
# but should not be found in the config file
NONCONFIG_SETTINGS = {
    'config_file': 'config',
    'settings_dir': user_config_dir('avendesora'),
    'default_accounts_file': 'accounts.gpg',
    'default_templates_file': 'templates',
    'charsets_hash': 'e4ae3714d9dbdffc0cf3b51a0462b5ec',
    'dict_hash': '11fe5bc734f4a956c37d7cb3da16ab3f',
    'secrets_hash': '74ad5101d77118a23d96aa29e2c695e9',
}


# Config Settings {{{1
# These are the default values for settings that may be found in the config file
CONFIG_DEFAULTS = {
    'accounts_files': [
        NONCONFIG_SETTINGS['default_accounts_file'],
        NONCONFIG_SETTINGS['default_templates_file'],
    ],
    'account_list_file': '.accounts_files',
    'archive_file': 'archive.gpg',
    'browsers': {
        'f': 'firefox -new-tab %s',
        'g': 'google-chrome %s',
        't': 'torbrowser %s',
        'x': 'xdg-open %s', # system default browser
    },
    'color_scheme': 'dark',
    'default_browser': 'x',
    'default_field': 'passcode',
    'default_vector_field': 'questions',
    'display_time': 60,
    'dictionary_file': 'words',
    'encoding': 'utf8',
    'edit_account': (
        'vim',
        '+silent /^class {account}(Account):/',
        '+silent normal zozt', # open the fold, position near top of screen
        '{filepath}'
    ),
    'edit_template': (
        'vim',
        '+silent /{field}/',
        '+silent normal zozt', # open the fold, position near top of screen
        '{filepath}'
    ),
    'hashes_file': '.hashes',
    'indent': '    ',
    'label_color': 'blue',
    'log_file': 'log.gpg',
    'required_protocols': ['https'],
    'user_key_file': '.key.gpg',
    'default_account_template': 'bank',
    'account_templates': {
        'website': """
            class _NAME_(Account): # %s1
                aliases = ['_ALIASES_']
                username = '_USERNAME_'
                email = '_EMAIL_'
                passcode = PasswordRecipe('12 2u 2d 2s')
                discovery = RecognizeURL(
                    'https://_URL_',
                    script='{email}{tab}{passcode}{return}'
                )
        """ % (3*'{'),
        'command': """
            class _NAME_(Account): # %s1
                aliases = ['_ALIASES_']
                passcode = Passphrase()
                discovery = RecognizeTitle(
                    '_TITLE1_', '_TITLE2_',
                    script='{passcode}{return}'
                )
        """ % (3*'{'),
        'bank': """
            class _NAME_(Account): # %s1
                aliases = ['_ALIASES_']
                username = '_NAME_'
                email = '_EMAIL_'
                accounts = {
                    'checking':   <<_ACCOUNT_NUMBER_>>,
                    'savings':    <<_ACCOUNT_NUMBER_>>,
                    'creditcard': <<_ACCOUNT_NUMBER_>>,
                }
                customer_service = '_PHONE_NUMBER_'
                passcode = PasswordRecipe('12 2u 2d 2s')
                verbal = Passphrase(length=2)
                pin = PIN()
                questions = [
                    Question('_QUESTION_'),
                    Question('_QUESTION_'),
                    Question('_QUESTION_'),
                ]
        """ % (3*'{'),
    },

    # use absolute paths for executables so they cannot be maliciously replaced
    # by changing the path.
    'gpg_executable': '/usr/bin/gpg2',
    'gpg_home': '~/.gnupg',
    'gpg_armor': 'extension',
    'gpg_ids': None,
    'xdotool_executable': '/usr/bin/xdotool',
    'xsel_executable': '/usr/bin/xsel',

}

# the following could be config settings, but they do not seem worth promoting
INITIAL_AUTOTYPE_DELAY = 0.0

# Initial config file {{{1
CONFIG_FILE_INITIAL_CONTENTS = dedent('''\
    # Include a few unicode characters, but to make sure they work: ±αβγδε
    # The desired location of the log file (relative to config directory).
    # Adding a suffix of .gpg or .asc causes the file to be encrypted
    # (otherwise it can leak account names). Use None to disable logging.
    log_file = {log_file}

    # The desired location of the archive file (relative to config director).
    # End the path in .gpg. Use None to disable archiving.
    archive_file = {archive_file}

    # Various settings
    default_field = {default_field}
    default_vector_field = {default_vector_field}
    display_time = {display_time}
    encoding = {encoding}
    edit_account = {edit_account}
    edit_template = {edit_template}
    browsers = {browsers}
    default_browser = {default_browser}
    required_protocols = {required_protocols}
    label_color = {label_color}
        # choose from black, red, green, yellow, blue, magenta, cyan, white
    color_scheme = {color_scheme}
        # choose from dark, light

    # Prototype accounts
    default_account_template = {default_account_template}
    account_templates = {account_templates}

    # Information used by GPG when encrypting and decrypting files.
    gpg_ids = {gpg_ids}
    gpg_armor = {gpg_armor}
        # choose between 'always', 'never' and 'by-ext' (.asc: armor, .gpg: no)
    gpg_home = {gpg_home}

    # Utilities
    # use of full absolute paths for executables is recommended
    gpg_executable = {gpg_executable}
    xdotool_executable = {xdotool_executable}
    xsel_executable = {xsel_executable}

    # vim: filetype=python sw=4 sts=4 et ai ff=unix nofen :
''')


# Initial hash file {{{1
HASH_FILE_INITIAL_CONTENTS = dedent('''\
    # Include a few unicode characters, but to make sure they work: ±αβγδε
    # Changing the contents of the dictionary, secrets, or charsets will change
    # the secrets you generate, thus you do not want to change these files once
    # you have started using the program. These are hashes for the contents of
    # these files at the time that this file was created. The program will
    # complain if the current contents of these files generate a different
    # hash. Only update these hashes if you know what you are doing.

    charsets_hash = {charsets_hash}  # DO NOT CHANGE THIS LINE
    dict_hash     = {dict_hash}  # DO NOT CHANGE THIS LINE
    secrets_hash  = {secrets_hash}  # DO NOT CHANGE THIS LINE
''')


# Initial accounts file {{{1
ACCOUNTS_FILE_INITIAL_CONTENTS = dedent('''\
    # Include a few unicode characters, but to make sure they work: ±αβγδε
    # Account information
    #
    # Add information about each of your accounts to the accounts dictionary.
    #
    # You can use the dedent function to strip leading whitespace from
    # multi-line remarks.  You can use the character sets and exclude function
    # to create alphabets for you character-base passwords.
    #
    # Example:
    # To create an alphabet with all characters except tabs use either:
    #     'alphabet': exclude(PRINTABLE, '\\t')
    # or:
    #     'alphabet': ALPHANUMERIC + PUNCTUATION + ' '
    #
    # vim: filetype=python sw=4 sts=4 et ai ff=unix :

    from avendesora import (
        # Basics
        Account, Hidden, GPG,

        # Character sets
        exclude, LOWERCASE, UPPERCASE, LETTERS, DIGITS, ALPHANUMERIC,
        HEXDIGITS, PUNCTUATION, SYMBOLS, WHITESPACE, PRINTABLE, DISTINGUISHABLE,

        # Secrets
        Password, Passphrase, PIN, Question, MixedPassword, PasswordRecipe,
        BirthDate,

        # Account Discovery
        RecognizeAll, RecognizeAny, RecognizeTitle, RecognizeURL, RecognizeCWD,
        RecognizeHost, RecognizeUser, RecognizeEnvVar,
    )

    master_password = Hidden({master_password})
    gpg_ids = {gpg_ids}

    # Accounts

''')

# Initial templates file {{{1
TEMPLATES_FILE_INITIAL_CONTENTS = dedent('''\
    # Include a few unicode characters, but to make sure they work: ±αβγδε
    # Templates
    #
    # Templates are accounts without master passwords and without much in the way of
    # attributes. Use them to generate pass codes of various types for stealth
    # accounts.

    # Imports {section}
    from avendesora import (
        # Character sets
        exclude, LOWERCASE, UPPERCASE, LETTERS, DIGITS, ALPHANUMERIC,
        HEXDIGITS, PUNCTUATION, WHITESPACE, PRINTABLE, DISTINGUISHABLE,

        # Account
        Account,

        # Secrets
        Password, Passphrase, PIN,
    )
    gpg_ids = {gpg_ids}

    # Templates
    # Alphanumeric passwords {section}
    class Anum4(Account):
        passcode = Password(length=4, alphabet=DISTINGUISHABLE)

    class Anum6(Account):
        passcode = Password(length=6, alphabet=DISTINGUISHABLE)

    class Anum8(Account):
        passcode = Password(length=8, alphabet=DISTINGUISHABLE)

    class Anum10(Account):
        passcode = Password(length=10, alphabet=DISTINGUISHABLE)

    class Anum12(Account):
        aliases = ['anum']
        passcode = Password(length=12, alphabet=DISTINGUISHABLE)

    class Anum14(Account):
        passcode = Password(length=14, alphabet=DISTINGUISHABLE)

    class Anum16(Account):
        passcode = Password(length=16, alphabet=DISTINGUISHABLE)

    class Anum18(Account):
        passcode = Password(length=18, alphabet=DISTINGUISHABLE)

    class Anum20(Account):
        passcode = Password(length=20, alphabet=DISTINGUISHABLE)

    # High entropy passwords {section}
    class Char4(Account):
        passcode = Password(length=4, alphabet=ALPHANUMERIC+PUNCTUATION)

    class Char6(Account):
        passcode = Password(length=6, alphabet=ALPHANUMERIC+PUNCTUATION)

    class Char8(Account):
        passcode = Password(length=8, alphabet=ALPHANUMERIC+PUNCTUATION)

    class Char10(Account):
        passcode = Password(length=10, alphabet=ALPHANUMERIC+PUNCTUATION)

    class Char12(Account):
        aliases = ['char']
        passcode = Password(length=12, alphabet=ALPHANUMERIC+PUNCTUATION)

    class Char14(Account):
        passcode = Password(length=14, alphabet=ALPHANUMERIC+PUNCTUATION)

    class Char16(Account):
        passcode = Password(length=16, alphabet=ALPHANUMERIC+PUNCTUATION)

    class Char18(Account):
        passcode = Password(length=18, alphabet=ALPHANUMERIC+PUNCTUATION)

    class Char20(Account):
        passcode = Password(length=20, alphabet=ALPHANUMERIC+PUNCTUATION)

    # Extreme passwords {section}
    class Extreme(Account):
        passcode = Passphrase(length=64, alphabet=PRINTABLE)

    # PINs {section}
    class Pin4(Account):
        aliases = ['pin']
        passcode = PIN(length=4)

    class Pin6(Account):
        passcode = PIN(length=6)

    class Pin8(Account):
        aliases = ['num']
        passcode = PIN(length=8)

    class Pin10(Account):
        passcode = PIN(length=10)

    class Pin12(Account):
        passcode = PIN(length=12)

    # Pass phrases {section}
    class Word(Account):
        passcode = Passphrase(length=1)

    class Word2(Account):
        aliases = ['pair']
        passcode = Passphrase(length=2)

    class Word3(Account):
        passcode = Passphrase(length=3)

    class Word4(Account):
        aliases = ['word', 'xkcd']
        passcode = Passphrase(length=4)

    class Word5(Account):
        passcode = Passphrase(length=5)

    class Word6(Account):
        passcode = Passphrase(length=6)

    class Word7(Account):
        passcode = Passphrase(length=7)

    class Word8(Account):
        passcode = Passphrase(length=8)

    # vim: filetype=python sw=4 sts=4 et ai ff=unix :
''')

# Initial user key file {{{1
USER_KEY_FILE_INITIAL_CONTENTS = dedent('''\
    # Include a few unicode characters, but to make sure they work: ±αβγδε
    # DO NOT CHANGE THIS KEY
    user_key = ({user_key})
''')

# Account list file {{{1
ACCOUNT_LIST_FILE_CONTENTS = dedent('''\
    # The list of files that contain accounts. The order of the files is
    # immaterial, except that first file given is the default file, meaning that
    # the add account will add a new account to the first files specified unless
    # a different files is specified explicitly.
    accounts_files = {accounts_files}
''')


# Fields {{{1
# Fields reserved for use by Avendesora
TOOL_FIELDS = ['aliases', 'default', 'master', 'discovery', 'browser']




# Player

# Imports {{{1

from scripts import (
    abspath, exists, extension, fopen, head, isdir, isfile, join, ls,
    normpath
)
from .prefs import mediaFileExtensions, skipSongThatWasPlayingWhenLastKilled
from .metadata import MetaData
from time import sleep
import sys
try:
    import _thread
except ImportError:
    import _thread
import gi
gi.require_version('Gst', '1.0')
from gi.repository import GObject, Gst
Gst.init(None)

# Player constructor{{{1
class Player(object):
    def __init__(self, quiet=False, now_playing_file = None):
        player = Gst.ElementFactory.make("playbin", "player")
        fakesink = Gst.ElementFactory.make("fakesink", "fakesink")
        player.set_property("video-sink", fakesink)
        bus = player.get_bus()
        bus.add_signal_watch()
        bus.connect("message", self.processMessage)
        self.player = player
        self.quiet = quiet
        self.now_playing_file = now_playing_file
        self.songs = []
        self.skip = []
        self.played = []
        self.playing = False

    # processMessage() {{{1
    def processMessage(self, bus, message):
        if message.type == Gst.MessageType.EOS:
            self.player.set_state(Gst.State.NULL)
            self.playing = False
        elif message.type == Gst.MessageType.ERROR:
            self.player.set_state(Gst.State.NULL)
            err, debug = message.parse_error()
            print("Error: %s" % err, debug)
            self.playing = False

    # addSongs() {{{1
    def addSongs(self, paths, cwd='.'):
        for path in paths:
            path = normpath(join(cwd, path))
            if isfile(path):
                ext = extension(path).lower()
                if ext in mediaFileExtensions:
                    self.songs += [path]
                elif ext == '.m3u':
                    with fopen(path) as playlist:
                        lines = [l.strip() for l in playlist.readlines()]
                        self.addSongs(
                            [l for l in lines if l and l[0] != '#'],
                            head(path)
                        )
                elif exists(path):
                    if not self.quiet:
                        #print("%s: skipping descriptor of unknown type." % path)
                        pass
                else:
                    if not self.quiet:
                        print("%s: no such file or directory." % path)
            elif isdir(path):
                self.addSongs(sorted(ls(path=path)))

    # writePlaylist() {{{1
    def writePlaylist(self, filename):
        with fopen(filename, 'w') as playlist:
            playlist.write('\n'.join(self.songs))

    # addSkips() {{{1
    def addSkips(self, paths):
        self.skip = paths

    # shuffleSongs() {{{1
    def shuffleSongs(self):
        from random import shuffle
        shuffle(self.songs)

    # play() {{{1
    def play(self, quit):
        for song in self.songs:
            if song in self.skip:
                continue
            if skipSongThatWasPlayingWhenLastKilled:
                self.played.append(song)
            self.playing = True
            if not self.quiet:
                metadata = MetaData(song, self.now_playing_file)
                print(metadata.summary())
                metadata.now_playing()
            self.player.set_property("uri", "file://" + abspath(song))
            self.player.set_state(Gst.State.PLAYING)
            while self.playing:
                sleep(0.1)
            if not skipSongThatWasPlayingWhenLastKilled:
                self.played.append(song)
        self.skip = []
        self.played = []
        sleep(1)
        quit()

    # songsAlreadyPlayed() {{{1
    def songsAlreadyPlayed(self):
        return self.skip + self.played
